"""Find out more at https://github.com/GDCC/pyDataverse.

Copyright 2019 Stefan Kasberger

Licensed under the MIT License.
"""

from __future__ import absolute_import


__author__ = "Stefan Kasberger"
__email__ = "stefan.kasberger@univie.ac.at"
__copyright__ = "Copyright (c) 2019 Stefan Kasberger"
__license__ = "MIT License"
__version__ = "0.3.5"
__url__ = "https://github.com/GDCC/pyDataverse"
__download_url__ = "https://pypi.python.org/pypi/pyDataverse"
__description__ = "A Python module for Dataverse."



"""Dataverse data-types data model."""

from __future__ import absolute_import

import json
import os

from pyDataverse.utils import validate_data


INTERNAL_ATTRIBUTES = [
    "_default_json_format",
    "_default_json_schema_filename",
    "_allowed_json_formats",
    "_json_dataverse_upload_attr",
    "_internal_attributes",
]


class DVObject:
    """Base class for the Dataverse data types `Dataverse`, `Dataset` and `Datafile`."""

    def __init__(self, data=None):
        """Init :class:`DVObject`.

        Parameters
        ----------
        data : dict
            Flat dictionary. All keys will be mapped to a similar
            named attribute and it's value.
        """
        if data is not None:
            self.set(data)

    def set(self, data):
        """Set class attributes by a flat dictionary.

        The flat dict is the main way to set the class attributes.
        It is the main interface between the object and the outside world.

        Parameters
        ----------
        data : dict
            Flat dictionary. All keys will be mapped to a similar
            named attribute and it's value.

        Returns
        -------
        bool
            `True` if all attributes are set, `False` if wrong data type was
            passed.

        """
        assert isinstance(data, dict)

        for key, val in data.items():
            if key in self._internal_attributes:
                print("Importing attribute {0} not allowed.".format(key))
            else:
                self.__setattr__(key, val)

    def get(self):
        """Create flat `dict` of all attributes.

        Creates :class:`dict` with all attributes in a flat structure.
        The flat :class:`dict` can then be used for further processing.

        Returns
        -------
        dict
            Data in a flat data structure.

        """
        data = {}

        for attr in list(self.__dict__.keys()):
            if attr not in INTERNAL_ATTRIBUTES:
                data[attr] = self.__getattribute__(attr)

        assert isinstance(data, dict)
        return data

    def validate_json(self, filename_schema=None):
        """Validate JSON formats.

        Check if JSON data structure is valid.

        Parameters
        ----------
        filename_schema : str
            Filename of JSON schema with full path.

        Returns
        -------
        bool
            `True` if JSON validates correctly, `False` if not.

        """
        if filename_schema is None:
            filename_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self._default_json_schema_filename,
            )
        assert isinstance(filename_schema, str)

        return validate_data(
            json.loads(self.json(validate=False)),
            filename_schema,
            file_format="json",
        )

    def from_json(
        self, json_str, data_format=None, validate=True, filename_schema=None
    ):
        """Import metadata from a JSON file.

        Parses in the metadata from different JSON formats.

        Parameters
        ----------
        json_str : str
            JSON string to be imported.
        data_format : str
            Data formats available for import. See `_allowed_json_formats`.
        validate : bool
            `True`, if imported JSON should be validated against a JSON
            schema file. `False`, if JSON string should be imported directly and
            not checked if valid.
        filename_schema : str
            Filename of JSON schema with full path.

        Returns
        -------
        bool
            `True` if JSON imported correctly, `False` if not.

        """
        assert isinstance(json_str, str)
        json_dict = json.loads(json_str)
        assert isinstance(json_dict, dict)
        assert isinstance(validate, bool)
        if data_format is None:
            data_format = self._default_json_format
        assert isinstance(data_format, str)
        assert data_format in self._allowed_json_formats
        if filename_schema is None:
            filename_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self._default_json_schema_filename,
            )
        assert isinstance(filename_schema, str)

        data = {}

        if data_format == "dataverse_upload":
            if validate:
                validate_data(json_dict, filename_schema)
            # get first level metadata and parse it automatically
            for key in json_dict.keys():
                if key in self._json_dataverse_upload_attr:
                    data[key] = json_dict[key]
                else:
                    print(
                        "INFO: Attribute {0} not valid for import (data format=`{1}`).".format(
                            key, data_format
                        )
                    )
        elif data_format == "dataverse_download":
            print("INFO: Not implemented yet.")
        elif data_format == "dspace":
            print("INFO: Not implemented yet.")
        elif data_format == "custom":
            print("INFO: Not implemented yet.")
        else:
            # TODO: add exception for wrong data format
            pass

        self.set(data)

    def json(self, data_format=None, validate=True, filename_schema=None):
        r"""Create JSON from :class:`DVObject` attributes.

        Parameters
        ----------
        data_format : str
            Data formats to be validated. See `_allowed_json_formats`.
        validate : bool
            `True`, if created JSON should be validated against a JSON schema
            file. `False`, if JSON string should be created and not checked if
            valid.
        filename_schema : str
            Filename of JSON schema with full path.

        Returns
        -------
        str
            The data as a JSON string.
        """
        assert isinstance(validate, bool)
        if data_format is None:
            data_format = self._default_json_format
        assert isinstance(data_format, str)
        assert data_format in self._allowed_json_formats
        if filename_schema is None:
            filename_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self._default_json_schema_filename,
            )
        assert isinstance(filename_schema, str)

        data = {}

        if data_format == "dataverse_upload":
            for attr in self._json_dataverse_upload_attr:
                # check if attribute exists
                if hasattr(self, attr):
                    data[attr] = self.__getattribute__(attr)
        elif data_format == "dspace":
            print("INFO: Not implemented yet.")
            return False
        elif data_format == "custom":
            print("INFO: Not implemented yet.")
            return False
        if validate:
            validate_data(data, filename_schema)

        json_str = json.dumps(data, indent=2)
        assert isinstance(json_str, str)
        return json_str


class Dataverse(DVObject):
    """Base class for the Dataverse data type `Dataverse`.

    Attributes
    ----------
    _default_json_format : str
        Default JSON data format.
    _default_json_schema_filename : str
        Default JSON schema filename.
    _allowed_json_formats : list
        List of all possible JSON data formats.
    _json_dataverse_upload_attr : list
        List of all attributes to be exported in :func:`json`.
    """

    def __init__(self, data=None):
        """Init :class:`Dataverse()`.

        Inherits attributes from parent :class:`DVObject()`

        Parameters
        ----------
        data : dict
            Flat dictionary. All keys will be mapped to a similar
            named attribute and it's value.

        Examples
        -------
        Create a Dataverse::

            >>> from pyDataverse.models import Dataverse
            >>> dv = Dataverse()
            >>> print(dv._default_json_schema_filename)
            'schemas/json/dataverse_upload_schema.json'

        """
        self._internal_attributes = [
            "_Dataverse" + attr for attr in INTERNAL_ATTRIBUTES
        ]

        super().__init__(data=data)

        self._default_json_format = "dataverse_upload"
        self._default_json_schema_filename = "schemas/json/dataverse_upload_schema.json"
        self._allowed_json_formats = ["dataverse_upload", "dataverse_download"]
        self._json_dataverse_upload_attr = [
            "affiliation",
            "alias",
            "dataverseContacts",
            "dataverseType",
            "description",
            "name",
        ]


class Dataset(DVObject):
    """Base class for the Dataverse data type `Dataset`.

    Attributes
    ----------
    _default_json_format : str
        Default JSON data format.
    _default_json_schema_filename : str
        Default JSON schema filename.
    _allowed_json_formats : list
        List of all possible JSON data formats.
    _json_dataverse_upload_attr : list
        List with all attributes to be exported in :func:`json`.
    __attr_import_dv_up_datasetVersion_values : list
        Dataverse API Upload Dataset JSON attributes inside ds[\'datasetVersion\'].
    __attr_import_dv_up_citation_fields_values : list
        Dataverse API Upload Dataset JSON attributes inside
        ds[\'datasetVersion\'][\'metadataBlocks\'][\'citation\'][\'fields\'].
    __attr_import_dv_up_citation_fields_arrays : dict
        Dataverse API Upload Dataset JSON attributes inside
        [\'datasetVersion\'][\'metadataBlocks\'][\'citation\'][\'fields\'].
    __attr_import_dv_up_geospatial_fields_values : list
        Attributes of Dataverse API Upload Dataset JSON metadata standard inside
        [\'datasetVersion\'][\'metadataBlocks\'][\'geospatial\'][\'fields\'].
    __attr_import_dv_up_geospatial_fields_arrays : dict
        Attributes of Dataverse API Upload Dataset JSON metadata standard inside
        [\'datasetVersion\'][\'metadataBlocks\'][\'geospatial\'][\'fields\'].
    __attr_import_dv_up_socialscience_fields_values : list
        Attributes of Dataverse API Upload Dataset JSON metadata standard inside
        [\'datasetVersion\'][\'metadataBlocks\'][\'socialscience\'][\'fields\'].
    __attr_import_dv_up_journal_fields_values : list
        Attributes of Dataverse API Upload Dataset JSON metadata standard inside
        [\'datasetVersion\'][\'metadataBlocks\'][\'journal\'][\'fields\'].
    __attr_import_dv_up_journal_fields_arrays : dict
        Attributes of Dataverse API Upload Dataset JSON metadata standard inside
        [\'datasetVersion\'][\'metadataBlocks\'][\'journal\'][\'fields\'].
    __attr_dict_dv_up_required :list
        Required attributes for valid `dv_up` metadata dict creation.
    __attr_dict_dv_up_type_class_primitive : list
        typeClass primitive.
    __attr_dict_dv_up_type_class_compound : list
        typeClass compound.
    __attr_dict_dv_up_type_class_controlled_vocabulary : list
        typeClass controlledVocabulary.
    __attr_dict_dv_up_single_dict : list
        This attributes are excluded from automatic parsing in ds.get() creation.
    __attr_displayNames : list
        Attributes of displayName.
    """

    __attr_import_dv_up_datasetVersion_values = [
        "license",
        "termsOfAccess",
        "fileAccessRequest",
        "protocol",
        "authority",
        "identifier",
        "termsOfUse",
    ]
    __attr_import_dv_up_citation_fields_values = [
        "accessToSources",
        "alternativeTitle",
        "alternativeURL",
        "characteristicOfSources",
        "dateOfDeposit",
        "dataSources",
        "depositor",
        "distributionDate",
        "kindOfData",
        "language",
        "notesText",
        "originOfSources",
        "otherReferences",
        "productionDate",
        "productionPlace",
        "relatedDatasets",
        "relatedMaterial",
        "subject",
        "subtitle",
        "title",
    ]
    __attr_import_dv_up_citation_fields_arrays = {
        "author": [
            "authorName",
            "authorAffiliation",
            "authorIdentifierScheme",
            "authorIdentifier",
        ],
        "contributor": ["contributorType", "contributorName"],
        "dateOfCollection": ["dateOfCollectionStart", "dateOfCollectionEnd"],
        "datasetContact": [
            "datasetContactName",
            "datasetContactAffiliation",
            "datasetContactEmail",
        ],
        "distributor": [
            "distributorName",
            "distributorAffiliation",
            "distributorAbbreviation",
            "distributorURL",
            "distributorLogoURL",
        ],
        "dsDescription": ["dsDescriptionValue", "dsDescriptionDate"],
        "grantNumber": ["grantNumberAgency", "grantNumberValue"],
        "keyword": ["keywordValue", "keywordVocabulary", "keywordVocabularyURI"],
        "producer": [
            "producerName",
            "producerAffiliation",
            "producerAbbreviation",
            "producerURL",
            "producerLogoURL",
        ],
        "otherId": ["otherIdAgency", "otherIdValue"],
        "publication": [
            "publicationCitation",
            "publicationIDType",
            "publicationIDNumber",
            "publicationURL",
        ],
        "software": ["softwareName", "softwareVersion"],
        "timePeriodCovered": ["timePeriodCoveredStart", "timePeriodCoveredEnd"],
        "topicClassification": [
            "topicClassValue",
            "topicClassVocab",
            "topicClassVocabURI",
        ],
    }
    __attr_import_dv_up_geospatial_fields_values = ["geographicUnit"]
    __attr_import_dv_up_geospatial_fields_arrays = {
        "geographicBoundingBox": [
            "westLongitude",
            "eastLongitude",
            "northLongitude",
            "southLongitude",
        ],
        "geographicCoverage": ["country", "state", "city", "otherGeographicCoverage"],
    }
    __attr_import_dv_up_socialscience_fields_values = [
        "actionsToMinimizeLoss",
        "cleaningOperations",
        "collectionMode",
        "collectorTraining",
        "controlOperations",
        "dataCollectionSituation",
        "dataCollector",
        "datasetLevelErrorNotes",
        "deviationsFromSampleDesign",
        "frequencyOfDataCollection",
        "otherDataAppraisal",
        "researchInstrument",
        "responseRate",
        "samplingErrorEstimates",
        "samplingProcedure",
        "unitOfAnalysis",
        "universe",
        "timeMethod",
        "weighting",
    ]
    __attr_import_dv_up_journal_fields_values = ["journalArticleType"]
    __attr_import_dv_up_journal_fields_arrays = {
        "journalVolumeIssue": ["journalVolume", "journalIssue", "journalPubDate"]
    }
    __attr_dict_dv_up_required = [
        "author",
        "datasetContact",
        "dsDescription",
        "subject",
        "title",
    ]
    __attr_dict_dv_up_type_class_primitive = (
        [
            "accessToSources",
            "alternativeTitle",
            "alternativeURL",
            "authorAffiliation",
            "authorIdentifier",
            "authorName",
            "characteristicOfSources",
            "city",
            "contributorName",
            "dateOfDeposit",
            "dataSources",
            "depositor",
            "distributionDate",
            "kindOfData",
            "notesText",
            "originOfSources",
            "otherGeographicCoverage",
            "otherReferences",
            "productionDate",
            "productionPlace",
            "publicationCitation",
            "publicationIDNumber",
            "publicationURL",
            "relatedDatasets",
            "relatedMaterial",
            "seriesInformation",
            "seriesName",
            "state",
            "subtitle",
            "title",
        ]
        + __attr_import_dv_up_citation_fields_arrays["dateOfCollection"]
        + __attr_import_dv_up_citation_fields_arrays["datasetContact"]
        + __attr_import_dv_up_citation_fields_arrays["distributor"]
        + __attr_import_dv_up_citation_fields_arrays["dsDescription"]
        + __attr_import_dv_up_citation_fields_arrays["grantNumber"]
        + __attr_import_dv_up_citation_fields_arrays["keyword"]
        + __attr_import_dv_up_citation_fields_arrays["producer"]
        + __attr_import_dv_up_citation_fields_arrays["otherId"]
        + __attr_import_dv_up_citation_fields_arrays["software"]
        + __attr_import_dv_up_citation_fields_arrays["timePeriodCovered"]
        + __attr_import_dv_up_citation_fields_arrays["topicClassification"]
        + __attr_import_dv_up_geospatial_fields_values
        + __attr_import_dv_up_geospatial_fields_arrays["geographicBoundingBox"]
        + __attr_import_dv_up_socialscience_fields_values
        + __attr_import_dv_up_journal_fields_arrays["journalVolumeIssue"]
        + [
            "socialScienceNotesType",
            "socialScienceNotesSubject",
            "socialScienceNotesText",
        ]
        + ["targetSampleActualSize", "targetSampleSizeFormula"]
    )
    __attr_dict_dv_up_type_class_compound = (
        list(__attr_import_dv_up_citation_fields_arrays.keys())
        + list(__attr_import_dv_up_geospatial_fields_arrays.keys())
        + list(__attr_import_dv_up_journal_fields_arrays.keys())
        + ["series", "socialScienceNotes", "targetSampleSize"]
    )
    __attr_dict_dv_up_type_class_controlled_vocabulary = [
        "authorIdentifierScheme",
        "contributorType",
        "country",
        "journalArticleType",
        "language",
        "publicationIDType",
        "subject",
    ]
    __attr_dict_dv_up_single_dict = ["series", "socialScienceNotes", "targetSampleSize"]
    __attr_displayNames = [
        "citation_displayName",
        "geospatial_displayName",
        "socialscience_displayName",
        "journal_displayName",
    ]

    def __init__(self, data=None):
        """Init a Dataset() class.

        Parameters
        ----------
        data : dict
            Flat dictionary. All keys will be mapped to a similar
            named attribute and it's value.

        Examples
        -------
        Create a Dataset::

            >>> from pyDataverse.models import Dataset
            >>> ds = Dataset()
            >>> print(ds._default_json_schema_filename)
            'schemas/json/dataset_upload_default_schema.json'

        """
        self._internal_attributes = ["_Dataset" + attr for attr in INTERNAL_ATTRIBUTES]

        super().__init__(data=data)

        self._default_json_format = "dataverse_upload"
        self._default_json_schema_filename = (
            "schemas/json/dataset_upload_default_schema.json"
        )
        self._allowed_json_formats = [
            "dataverse_upload",
            "dataverse_download",
            "dspace",
            "custom",
        ]
        self._json_dataverse_upload_attr = [
            "license",
            "termsOfUse",
            "termsOfAccess",
            "fileAccessRequest",
            "protocol",
            "authority",
            "identifier",
            "citation_displayName",
            "title",
            "subtitle",
            "alternativeTitle",
            "alternativeURL",
            "otherId",
            "author",
            "datasetContact",
            "dsDescription",
            "subject",
            "keyword",
            "topicClassification",
            "publication",
            "notesText",
            "producer",
            "productionDate",
            "productionPlace",
            "contributor",
            "grantNumber",
            "distributor",
            "distributionDate",
            "depositor",
            "dateOfDeposit",
            "timePeriodCovered",
            "dateOfCollection",
            "kindOfData",
            "language",
            "series",
            "software",
            "relatedMaterial",
            "relatedDatasets",
            "otherReferences",
            "dataSources",
            "originOfSources",
            "characteristicOfSources",
            "accessToSources",
            "geospatial_displayName",
            "geographicCoverage",
            "geographicUnit",
            "geographicBoundingBox",
            "socialscience_displayName",
            "unitOfAnalysis",
            "universe",
            "timeMethod",
            "dataCollector",
            "collectorTraining",
            "frequencyOfDataCollection",
            "samplingProcedure",
            "targetSampleSize",
            "deviationsFromSampleDesign",
            "collectionMode",
            "researchInstrument",
            "dataCollectionSituation",
            "actionsToMinimizeLoss",
            "controlOperations",
            "weighting",
            "cleaningOperations",
            "datasetLevelErrorNotes",
            "responseRate",
            "samplingErrorEstimates",
            "otherDataAppraisal",
            "socialScienceNotes",
            "journal_displayName",
            "journalVolumeIssue",
            "journalArticleType",
        ]

    def validate_json(self, filename_schema=None):
        """Validate JSON formats of Dataset.

        Check if JSON data structure is valid.

        Parameters
        ----------
        filename_schema : str
            Filename of JSON schema with full path.

        Returns
        -------
        bool
            `True` if JSON validate correctly, `False` if not.

        Examples
        -------
        Check if JSON is valid for Dataverse API upload::

            >>> from pyDataverse.models import Dataset
            >>> ds = Dataset()
            >>> data = {
            >>>     'title': 'pyDataverse study 2019',
            >>>     'dsDescription': [
            >>>         {'dsDescriptionValue': 'New study about pyDataverse usage in 2019'}
            >>>     ]
            >>> }
            >>> ds.set(data)
            >>> print(ds.validate_json())
            False
            >>> ds.author = [{'authorName': 'LastAuthor1, FirstAuthor1'}]
            >>> ds.datasetContact = [{'datasetContactName': 'LastContact1, FirstContact1'}]
            >>> ds.subject = ['Engineering']
            >>> print(ds.validate_json())
            True

        """
        if filename_schema is None:
            filename_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self._default_json_schema_filename,
            )
        assert isinstance(filename_schema, str)

        is_valid = True

        data_json = self.json(validate=False)
        if data_json:
            is_valid = validate_data(
                json.loads(data_json), filename_schema, file_format="json"
            )
            if not is_valid:
                return False
        else:
            return False

        # check if all required attributes are set
        for attr in self.__attr_dict_dv_up_required:
            if attr in list(self.__dict__.keys()):
                if not self.__getattribute__(attr):
                    is_valid = False
                    print("Attribute '{0}' is `False`.".format(attr))
            else:
                is_valid = False
                print("Attribute '{0}' missing.".format(attr))

        # check if attributes set are complete where necessary
        if "timePeriodCovered" in list(self.__dict__.keys()):
            tp_cov = self.__getattribute__("timePeriodCovered")
            if tp_cov:
                for tp in tp_cov:
                    if "timePeriodCoveredStart" in tp or "timePeriodCoveredEnd" in tp:
                        if not (
                            "timePeriodCoveredStart" in tp
                            and "timePeriodCoveredEnd" in tp
                        ):
                            is_valid = False
                            print("timePeriodCovered attribute missing.")

        if "dateOfCollection" in list(self.__dict__.keys()):
            d_coll = self.__getattribute__("dateOfCollection")
            if d_coll:
                for d in d_coll:
                    if "dateOfCollectionStart" in d or "dateOfCollectionEnd" in d:
                        if not (
                            "dateOfCollectionStart" in d and "dateOfCollectionEnd" in d
                        ):
                            is_valid = False
                            print("dateOfCollection attribute missing.")

        if "author" in list(self.__dict__.keys()):
            authors = self.__getattribute__("author")
            if authors:
                for a in authors:
                    if (
                        "authorAffiliation" in a
                        or "authorIdentifierScheme" in a
                        or "authorIdentifier" in a
                    ):
                        if "authorName" not in a:
                            is_valid = False
                            print("author attribute missing.")

        if "datasetContact" in list(self.__dict__.keys()):
            ds_contac = self.__getattribute__("datasetContact")
            if ds_contac:
                for c in ds_contac:
                    if "datasetContactAffiliation" in c or "datasetContactEmail" in c:
                        if "datasetContactName" not in c:
                            is_valid = False
                            print("datasetContact attribute missing.")

        if "producer" in list(self.__dict__.keys()):
            producer = self.__getattribute__("producer")
            if producer:
                for p in producer:
                    if (
                        "producerAffiliation" in p
                        or "producerAbbreviation" in p
                        or "producerURL" in p
                        or "producerLogoURL" in p
                    ):
                        if not p["producerName"]:
                            is_valid = False
                            print("producer attribute missing.")

        if "contributor" in list(self.__dict__.keys()):
            contributor = self.__getattribute__("contributor")
            if contributor:
                for c in contributor:
                    if "contributorType" in c:
                        if "contributorName" not in c:
                            is_valid = False
                            print("contributor attribute missing.")

        if "distributor" in list(self.__dict__.keys()):
            distributor = self.__getattribute__("distributor")
            if distributor:
                for d in distributor:
                    if (
                        "distributorAffiliation" in d
                        or "distributorAbbreviation" in d
                        or "distributorURL" in d
                        or "distributorLogoURL" in d
                    ):
                        if "distributorName" not in d:
                            is_valid = False
                            print("distributor attribute missing.")

        if "geographicBoundingBox" in list(self.__dict__.keys()):
            bbox = self.__getattribute__("geographicBoundingBox")
            if bbox:
                for b in bbox:
                    if b:
                        if not (
                            "westLongitude" in b
                            and "eastLongitude" in b
                            and "northLongitude" in b
                            and "southLongitude" in b
                        ):
                            is_valid = False
                            print("geographicBoundingBox attribute missing.")

        assert isinstance(is_valid, bool)
        return is_valid

    def from_json(
        self, json_str, data_format=None, validate=True, filename_schema=None
    ):
        """Import Dataset metadata from JSON file.

        Parses in the metadata of a Dataset from different JSON formats.

        Parameters
        ----------
        json_str : str
            JSON string to be imported.
        data_format : str
            Data formats available for import. See `_allowed_json_formats`.
        validate : bool
            `True`, if imported JSON should be validated against a JSON
            schema file. `False`, if JSON string should be imported directly and
            not checked if valid.
        filename_schema : str
            Filename of JSON schema with full path.

        Examples
        -------
        Set Dataverse attributes via flat :class:`dict`::

            >>> from pyDataverse.models import Dataset
            >>> ds = Dataset()
            >>> ds.from_json('tests/data/dataset_upload_min_default.json')
            >>> ds.title
            'Darwin's Finches'

        """
        assert isinstance(json_str, str)
        json_dict = json.loads(json_str)
        assert isinstance(json_dict, dict)
        assert isinstance(validate, bool)
        if data_format is None:
            data_format = self._default_json_format
        assert isinstance(data_format, str)
        assert data_format in self._allowed_json_formats
        if filename_schema is None:
            filename_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self._default_json_schema_filename,
            )
        assert isinstance(filename_schema, str)

        data = {}

        if data_format == "dataverse_upload":
            if validate:
                validate_data(json_dict, filename_schema, file_format="json")
            # dataset
            # get first level metadata and parse it automatically
            for key, val in json_dict["datasetVersion"].items():
                if not key == "metadataBlocks":
                    if key in self.__attr_import_dv_up_datasetVersion_values:
                        data[key] = val
                    else:
                        print(
                            "Attribute {0} not valid for import (format={1}).".format(
                                key, data_format
                            )
                        )

            if "metadataBlocks" in json_dict["datasetVersion"]:
                # citation
                if "citation" in json_dict["datasetVersion"]["metadataBlocks"]:
                    citation = json_dict["datasetVersion"]["metadataBlocks"]["citation"]
                    if "displayName" in citation:
                        data["citation_displayName"] = citation["displayName"]

                    for field in citation["fields"]:
                        if (
                            field["typeName"]
                            in self.__attr_import_dv_up_citation_fields_values
                        ):
                            data[field["typeName"]] = field["value"]
                        elif (
                            field["typeName"]
                            in self.__attr_import_dv_up_citation_fields_arrays
                        ):
                            data[field["typeName"]] = self.__parse_field_array(
                                field["value"],
                                self.__attr_import_dv_up_citation_fields_arrays[
                                    field["typeName"]
                                ],
                            )
                        elif field["typeName"] == "series":
                            data["series"] = {}
                            if "seriesName" in field["value"]:
                                data["series"]["seriesName"] = field["value"][
                                    "seriesName"
                                ]["value"]
                            if "seriesInformation" in field["value"]:
                                data["series"]["seriesInformation"] = field["value"][
                                    "seriesInformation"
                                ]["value"]
                        else:
                            print(
                                "Attribute {0} not valid for import (dv_up).".format(
                                    field["typeName"]
                                )
                            )
                else:
                    # TODO: Exception
                    pass

                # geospatial
                if "geospatial" in json_dict["datasetVersion"]["metadataBlocks"]:
                    geospatial = json_dict["datasetVersion"]["metadataBlocks"][
                        "geospatial"
                    ]
                    if "displayName" in geospatial:
                        self.__setattr__(
                            "geospatial_displayName", geospatial["displayName"]
                        )

                    for field in geospatial["fields"]:
                        if (
                            field["typeName"]
                            in self.__attr_import_dv_up_geospatial_fields_values
                        ):
                            data[field["typeName"]] = field["value"]
                        elif (
                            field["typeName"]
                            in self.__attr_import_dv_up_geospatial_fields_arrays
                        ):
                            data[field["typeName"]] = self.__parse_field_array(
                                field["value"],
                                self.__attr_import_dv_up_geospatial_fields_arrays[
                                    field["typeName"]
                                ],
                            )
                        else:
                            print(
                                "Attribute {0} not valid for import (dv_up).".format(
                                    field["typeName"]
                                )
                            )
                else:
                    # TODO: Exception
                    pass

                # socialscience
                if "socialscience" in json_dict["datasetVersion"]["metadataBlocks"]:
                    socialscience = json_dict["datasetVersion"]["metadataBlocks"][
                        "socialscience"
                    ]

                    if "displayName" in socialscience:
                        self.__setattr__(
                            "socialscience_displayName",
                            socialscience["displayName"],
                        )

                    for field in socialscience["fields"]:
                        if (
                            field["typeName"]
                            in self.__attr_import_dv_up_socialscience_fields_values
                        ):
                            data[field["typeName"]] = field["value"]
                        elif field["typeName"] == "targetSampleSize":
                            data["targetSampleSize"] = {}
                            if "targetSampleActualSize" in field["value"]:
                                data["targetSampleSize"]["targetSampleActualSize"] = (
                                    field["value"]["targetSampleActualSize"]["value"]
                                )
                            if "targetSampleSizeFormula" in field["value"]:
                                data["targetSampleSize"]["targetSampleSizeFormula"] = (
                                    field["value"]["targetSampleSizeFormula"]["value"]
                                )
                        elif field["typeName"] == "socialScienceNotes":
                            data["socialScienceNotes"] = {}
                            if "socialScienceNotesType" in field["value"]:
                                data["socialScienceNotes"]["socialScienceNotesType"] = (
                                    field["value"]["socialScienceNotesType"]["value"]
                                )
                            if "socialScienceNotesSubject" in field["value"]:
                                data["socialScienceNotes"][
                                    "socialScienceNotesSubject"
                                ] = field["value"]["socialScienceNotesSubject"]["value"]
                            if "socialScienceNotesText" in field["value"]:
                                data["socialScienceNotes"]["socialScienceNotesText"] = (
                                    field["value"]["socialScienceNotesText"]["value"]
                                )
                        else:
                            print(
                                "Attribute {0} not valid for import (dv_up).".format(
                                    field["typeName"]
                                )
                            )
                else:
                    # TODO: Exception
                    pass

                # journal
                if "journal" in json_dict["datasetVersion"]["metadataBlocks"]:
                    journal = json_dict["datasetVersion"]["metadataBlocks"]["journal"]

                    if "displayName" in journal:
                        self.__setattr__("journal_displayName", journal["displayName"])

                    for field in journal["fields"]:
                        if (
                            field["typeName"]
                            in self.__attr_import_dv_up_journal_fields_values
                        ):
                            data[field["typeName"]] = field["value"]
                        elif (
                            field["typeName"]
                            in self.__attr_import_dv_up_journal_fields_arrays
                        ):
                            data[field["typeName"]] = self.__parse_field_array(
                                field["value"],
                                self.__attr_import_dv_up_journal_fields_arrays[
                                    field["typeName"]
                                ],
                            )
                        else:
                            print(
                                "Attribute {0} not valid for import (dv_up).".format(
                                    field["typeName"]
                                )
                            )
                else:
                    # TODO: Exception
                    pass
        elif data_format == "dataverse_download":
            print("INFO: Not implemented yet.")
        elif data_format == "dspace":
            print("INFO: Not implemented yet.")
        elif data_format == "custom":
            print("INFO: Not implemented yet.")
        self.set(data)

    def __parse_field_array(self, data, attr_list):
        """Parse arrays of Dataset upload format.

        Parameters
        ----------
        data : list
            List of dictionaries of a specific Dataverse API metadata field.
        attr_list : list
            List of attributes to be parsed.

        Returns
        -------
        list
            List of :class:`dict`s with parsed out key-value pairs.

        """
        assert isinstance(data, list)
        assert isinstance(attr_list, list)

        data_tmp = []

        for d in data:
            tmp_dict = {}
            for key, val in d.items():
                if key in attr_list:
                    tmp_dict[key] = val["value"]
                else:
                    print("Key '{0}' not in attribute list".format(key))
            data_tmp.append(tmp_dict)

        assert isinstance(data_tmp, list)
        return data_tmp

    def __generate_field_arrays(self, key, sub_keys):
        """Generate dicts for array attributes of Dataverse API metadata upload.

        Parameters
        ----------
        key : str
            Name of attribute.
        sub_keys : list
            List of keys to be created.

        Returns
        -------
        list
            List of filled :class:`dict`s of metadata for Dataverse API upload.

        """
        assert isinstance(key, str)
        assert isinstance(sub_keys, list)

        # check if attribute exists
        tmp_list = []
        if self.__getattribute__(key):
            attr = self.__getattribute__(key)
            # loop over list of attribute dict
            for d in attr:
                tmp_dict = {}
                # iterate over key-value pairs
                for k, v in d.items():
                    # check if key is in attribute list
                    if k in sub_keys:
                        multiple = None
                        type_class = None
                        if isinstance(v, list):
                            multiple = True
                        else:
                            multiple = False
                        if k in self.__attr_dict_dv_up_type_class_primitive:
                            type_class = "primitive"
                        elif k in self.__attr_dict_dv_up_type_class_compound:
                            type_class = "compound"
                        elif (
                            k in self.__attr_dict_dv_up_type_class_controlled_vocabulary
                        ):
                            type_class = "controlledVocabulary"
                        tmp_dict[k] = {}
                        tmp_dict[k]["typeName"] = k
                        tmp_dict[k]["typeClass"] = type_class
                        tmp_dict[k]["multiple"] = multiple
                        tmp_dict[k]["value"] = v
                tmp_list.append(tmp_dict)

        assert isinstance(tmp_list, list)
        return tmp_list

    def json(self, data_format=None, validate=True, filename_schema=None):
        """Create Dataset JSON from attributes.

        Parameters
        ----------
        format : str
            Data formats to be validated. See `_allowed_json_formats`.
        validate : bool
            `True`, if created JSON should be validated against a JSON schema
            file. `False`, if JSON string should be created and not checked if
            valid.
        filename_schema : str
            Filename of JSON schema with full path.

        Returns
        -------
        str
            The data as a JSON string.
        """
        assert isinstance(validate, bool)
        if data_format is None:
            data_format = self._default_json_format
        assert isinstance(data_format, str)
        assert data_format in self._allowed_json_formats
        if filename_schema is None:
            filename_schema = os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                self._default_json_schema_filename,
            )
        assert isinstance(filename_schema, str)

        data = {}

        if data_format == "dataverse_upload":
            data_dict = self.get()
            data["datasetVersion"] = {}
            data["datasetVersion"]["metadataBlocks"] = {}
            citation = {}
            citation["fields"] = []

            # dataset
            # Generate first level attributes
            for attr in self.__attr_import_dv_up_datasetVersion_values:
                if attr in data_dict:
                    data["datasetVersion"][attr] = data_dict[attr]

            # citation
            if "citation_displayName" in data_dict:
                citation["displayName"] = data_dict["citation_displayName"]

            # Generate first level attributes
            for attr in self.__attr_import_dv_up_citation_fields_values:
                if attr in data_dict:
                    v = data_dict[attr]
                    if isinstance(v, list):
                        multiple = True
                    else:
                        multiple = False
                    if attr in self.__attr_dict_dv_up_type_class_primitive:
                        type_class = "primitive"
                    elif attr in self.__attr_dict_dv_up_type_class_compound:
                        type_class = "compound"
                    elif (
                        attr in self.__attr_dict_dv_up_type_class_controlled_vocabulary
                    ):
                        type_class = "controlledVocabulary"
                    citation["fields"].append(
                        {
                            "typeName": attr,
                            "multiple": multiple,
                            "typeClass": type_class,
                            "value": v,
                        }
                    )

            # Generate fields attributes
            for (
                key,
                val,
            ) in self.__attr_import_dv_up_citation_fields_arrays.items():
                if key in data_dict:
                    v = data_dict[key]
                    citation["fields"].append(
                        {
                            "typeName": key,
                            "multiple": True,
                            "typeClass": "compound",
                            "value": self.__generate_field_arrays(key, val),
                        }
                    )

            # Generate series attributes
            if "series" in data_dict:
                series = data_dict["series"]
                tmp_dict = {}
                if "seriesName" in series:
                    if series["seriesName"] is not None:
                        tmp_dict["seriesName"] = {}
                        tmp_dict["seriesName"]["typeName"] = "seriesName"
                        tmp_dict["seriesName"]["multiple"] = False
                        tmp_dict["seriesName"]["typeClass"] = "primitive"
                        tmp_dict["seriesName"]["value"] = series["seriesName"]
                if "seriesInformation" in series:
                    if series["seriesInformation"] is not None:
                        tmp_dict["seriesInformation"] = {}
                        tmp_dict["seriesInformation"]["typeName"] = "seriesInformation"
                        tmp_dict["seriesInformation"]["multiple"] = False
                        tmp_dict["seriesInformation"]["typeClass"] = "primitive"
                        tmp_dict["seriesInformation"]["value"] = series[
                            "seriesInformation"
                        ]
                citation["fields"].append(
                    {
                        "typeName": "series",
                        "multiple": False,
                        "typeClass": "compound",
                        "value": tmp_dict,
                    }
                )

            # geospatial
            for attr in (
                self.__attr_import_dv_up_geospatial_fields_values
                + list(self.__attr_import_dv_up_geospatial_fields_arrays.keys())
                + ["geospatial_displayName"]
            ):
                if attr in data_dict:
                    geospatial = {}
                    if attr != "geospatial_displayName":
                        geospatial["fields"] = []
                        break

            if "geospatial_displayName" in data_dict:
                geospatial["displayName"] = data_dict["geospatial_displayName"]

            # Generate first level attributes
            for attr in self.__attr_import_dv_up_geospatial_fields_values:
                if attr in data_dict:
                    v = data_dict[attr]
                    if isinstance(v, list):
                        multiple = True
                    else:
                        multiple = False
                    if attr in self.__attr_dict_dv_up_type_class_primitive:
                        type_class = "primitive"
                    elif attr in self.__attr_dict_dv_up_type_class_compound:
                        type_class = "compound"
                    elif (
                        attr in self.__attr_dict_dv_up_type_class_controlled_vocabulary
                    ):
                        type_class = "controlledVocabulary"
                    geospatial["fields"].append(
                        {
                            "typeName": attr,
                            "multiple": multiple,
                            "typeClass": type_class,
                            "value": v,
                        }
                    )

            # Generate fields attributes
            for (
                key,
                val,
            ) in self.__attr_import_dv_up_geospatial_fields_arrays.items():
                if key in data_dict:
                    geospatial["fields"].append(
                        {
                            "typeName": key,
                            "multiple": True,
                            "typeClass": "compound",
                            "value": self.__generate_field_arrays(key, val),
                        }
                    )

            # socialscience
            for attr in self.__attr_import_dv_up_socialscience_fields_values + [
                "socialscience_displayName"
            ]:
                if attr in data_dict:
                    socialscience = {}
                    if attr != "socialscience_displayName":
                        socialscience["fields"] = []
                        break

            if "socialscience_displayName" in data_dict:
                socialscience["displayName"] = data_dict["socialscience_displayName"]

            # Generate first level attributes
            for attr in self.__attr_import_dv_up_socialscience_fields_values:
                if attr in data_dict:
                    v = data_dict[attr]
                    if isinstance(v, list):
                        multiple = True
                    else:
                        multiple = False
                    if attr in self.__attr_dict_dv_up_type_class_primitive:
                        type_class = "primitive"
                    elif attr in self.__attr_dict_dv_up_type_class_compound:
                        type_class = "compound"
                    elif (
                        attr in self.__attr_dict_dv_up_type_class_controlled_vocabulary
                    ):
                        type_class = "controlledVocabulary"
                    socialscience["fields"].append(
                        {
                            "typeName": attr,
                            "multiple": multiple,
                            "typeClass": type_class,
                            "value": v,
                        }
                    )

            # Generate targetSampleSize attributes
            if "targetSampleSize" in data_dict:
                target_sample_size = data_dict["targetSampleSize"]
                tmp_dict = {}
                if "targetSampleActualSize" in target_sample_size:
                    if target_sample_size["targetSampleActualSize"] is not None:
                        tmp_dict["targetSampleActualSize"] = {}
                        tmp_dict["targetSampleActualSize"]["typeName"] = (
                            "targetSampleActualSize"
                        )
                        tmp_dict["targetSampleActualSize"]["multiple"] = False
                        tmp_dict["targetSampleActualSize"]["typeClass"] = "primitive"
                        tmp_dict["targetSampleActualSize"]["value"] = (
                            target_sample_size["targetSampleActualSize"]
                        )
                if "targetSampleSizeFormula" in target_sample_size:
                    if target_sample_size["targetSampleSizeFormula"] is not None:
                        tmp_dict["targetSampleSizeFormula"] = {}
                        tmp_dict["targetSampleSizeFormula"]["typeName"] = (
                            "targetSampleSizeFormula"
                        )
                        tmp_dict["targetSampleSizeFormula"]["multiple"] = False
                        tmp_dict["targetSampleSizeFormula"]["typeClass"] = "primitive"
                        tmp_dict["targetSampleSizeFormula"]["value"] = (
                            target_sample_size["targetSampleSizeFormula"]
                        )
                socialscience["fields"].append(
                    {
                        "typeName": "targetSampleSize",
                        "multiple": False,
                        "typeClass": "compound",
                        "value": tmp_dict,
                    }
                )

            # Generate socialScienceNotes attributes
            if "socialScienceNotes" in data_dict:
                social_science_notes = data_dict["socialScienceNotes"]
                tmp_dict = {}
                if "socialScienceNotesType" in social_science_notes:
                    if social_science_notes["socialScienceNotesType"] is not None:
                        tmp_dict["socialScienceNotesType"] = {}
                        tmp_dict["socialScienceNotesType"]["typeName"] = (
                            "socialScienceNotesType"
                        )
                        tmp_dict["socialScienceNotesType"]["multiple"] = False
                        tmp_dict["socialScienceNotesType"]["typeClass"] = "primitive"
                        tmp_dict["socialScienceNotesType"]["value"] = (
                            social_science_notes["socialScienceNotesType"]
                        )
                if "socialScienceNotesSubject" in social_science_notes:
                    if social_science_notes["socialScienceNotesSubject"] is not None:
                        tmp_dict["socialScienceNotesSubject"] = {}
                        tmp_dict["socialScienceNotesSubject"]["typeName"] = (
                            "socialScienceNotesSubject"
                        )
                        tmp_dict["socialScienceNotesSubject"]["multiple"] = False
                        tmp_dict["socialScienceNotesSubject"]["typeClass"] = "primitive"
                        tmp_dict["socialScienceNotesSubject"]["value"] = (
                            social_science_notes["socialScienceNotesSubject"]
                        )
                if "socialScienceNotesText" in social_science_notes:
                    if social_science_notes["socialScienceNotesText"] is not None:
                        tmp_dict["socialScienceNotesText"] = {}
                        tmp_dict["socialScienceNotesText"]["typeName"] = (
                            "socialScienceNotesText"
                        )
                        tmp_dict["socialScienceNotesText"]["multiple"] = False
                        tmp_dict["socialScienceNotesText"]["typeClass"] = "primitive"
                        tmp_dict["socialScienceNotesText"]["value"] = (
                            social_science_notes["socialScienceNotesText"]
                        )
                socialscience["fields"].append(
                    {
                        "typeName": "socialScienceNotes",
                        "multiple": False,
                        "typeClass": "compound",
                        "value": tmp_dict,
                    }
                )

            # journal
            for attr in (
                self.__attr_import_dv_up_journal_fields_values
                + list(self.__attr_import_dv_up_journal_fields_arrays.keys())
                + ["journal_displayName"]
            ):
                if attr in data_dict:
                    journal = {}
                    if attr != "journal_displayName":
                        journal["fields"] = []
                        break

            if "journal_displayName" in data_dict:
                journal["displayName"] = data_dict["journal_displayName"]

            # Generate first level attributes
            for attr in self.__attr_import_dv_up_journal_fields_values:
                if attr in data_dict:
                    v = data_dict[attr]
                    if isinstance(v, list):
                        multiple = True
                    else:
                        multiple = False
                    if attr in self.__attr_dict_dv_up_type_class_primitive:
                        type_class = "primitive"
                    elif attr in self.__attr_dict_dv_up_type_class_compound:
                        type_class = "compound"
                    elif (
                        attr in self.__attr_dict_dv_up_type_class_controlled_vocabulary
                    ):
                        type_class = "controlledVocabulary"
                    journal["fields"].append(
                        {
                            "typeName": attr,
                            "multiple": multiple,
                            "typeClass": type_class,
                            "value": v,
                        }
                    )

            # Generate fields attributes
            for (
                key,
                val,
            ) in self.__attr_import_dv_up_journal_fields_arrays.items():
                if key in data_dict:
                    journal["fields"].append(
                        {
                            "typeName": key,
                            "multiple": True,
                            "typeClass": "compound",
                            "value": self.__generate_field_arrays(key, val),
                        }
                    )

            data["datasetVersion"]["metadataBlocks"]["citation"] = citation
            if "socialscience" in locals():
                data["datasetVersion"]["metadataBlocks"]["socialscience"] = (
                    socialscience
                )
            if "geospatial" in locals():
                data["datasetVersion"]["metadataBlocks"]["geospatial"] = geospatial
            if "journal" in locals():
                data["datasetVersion"]["metadataBlocks"]["journal"] = journal
        elif data_format == "dspace":
            data = None
            print("INFO: Not implemented yet.")
        elif data_format == "custom":
            data = None
            print("INFO: Not implemented yet.")
        if validate:
            validate_data(data, filename_schema)

        json_str = json.dumps(data, indent=2)
        assert isinstance(json_str, str)
        return json_str


class Datafile(DVObject):
    """Base class for the Dataverse data type `Datafile`.

    Attributes
    ----------
    _default_json_format : str
        Default JSON data format.
    _default_json_schema_filename : str
        Default JSON schema filename.
    _allowed_json_formats : list
        List of all possible JSON data formats.
    _json_dataverse_upload_attr : list
        List of all attributes to be exported in :func:`json`.
    """

    def __init__(self, data=None):
        """Init :class:`Datafile()`.

        Inherits attributes from parent :class:`DVObject()`

        Parameters
        ----------
        data : dict
            Flat dictionary. All keys will be mapped to a similar
            named attribute and it's value.

        Examples
        -------
        Create a Datafile::

            >>> from pyDataverse.models import Datafile
            >>> df = Datafile()
            >>> print(df._default_json_schema_filename)
            'schemas/json/datafile_upload_schema.json'

        """
        self._internal_attributes = ["_Datafile" + attr for attr in INTERNAL_ATTRIBUTES]

        super().__init__(data=data)

        self._default_json_format = "dataverse_upload"
        self._default_json_schema_filename = "schemas/json/datafile_upload_schema.json"
        self._allowed_json_formats = ["dataverse_upload", "dataverse_download"]
        self._json_dataverse_upload_attr = [
            "description",
            "categories",
            "restrict",
            "label",
            "directoryLabel",
            "pid",
            "filename",
        ]



"""This module contains authentication handlers compatible with :class:`httpx.Auth`"""

from typing import Generator

from httpx import Auth, Request, Response

from pyDataverse.exceptions import ApiAuthorizationError


class ApiTokenAuth(Auth):
    """An authentication handler to add an API token as the X-Dataverse-key
    header.

    For more information on how to retrieve an API token and how it is used,
    please refer to https://guides.dataverse.org/en/latest/api/auth.html.
    """

    def __init__(self, api_token: str):
        """Initializes the auth handler with an API token.

        Parameters
        ----------
        api_token : str
            The API token retrieved from your Dataverse instance user profile.

        Examples
        --------

            >>> import os
            >>> from pyDataverse.api import DataAccessApi
            >>> base_url = 'https://demo.dataverse.org'
            >>> api_token_auth = ApiTokenAuth(os.getenv('API_TOKEN'))
            >>> api = DataAccessApi(base_url, api_token_auth)

        """
        if not isinstance(api_token, str):
            raise ApiAuthorizationError("API token passed is not a string.")
        self.api_token = api_token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Adds the X-Dataverse-key header with the API token and yields the
        original :class:`httpx.Request`.

        Parameters
        ----------
        request : httpx.Request
            The request object which requires authentication headers

        Yields
        ------
        httpx.Request
            The original request with modified headers
        """
        request.headers["X-Dataverse-key"] = self.api_token
        yield request


class BearerTokenAuth(Auth):
    """An authentication handler to add a Bearer token as defined in `RFC 6750
    <https://datatracker.ietf.org/doc/html/rfc6750>`_ to the request.

    A bearer token could be obtained from an OIDC provider, for example,
    Keycloak.
    """

    def __init__(self, bearer_token: str):
        """Initializes the auth handler with a bearer token.

        Parameters
        ----------
        bearer_token : str
            The bearer token retrieved from your OIDC provider.

        Examples
        --------

            >>> import os
            >>> from pyDataverse.api import DataAccessApi
            >>> base_url = 'https://demo.dataverse.org'
            >>> bearer_token_auth = OAuthBearerTokenAuth(os.getenv('OAUTH_TOKEN'))
            >>> api = DataAccessApi(base_url, bearer_token_auth)

        """
        if not isinstance(bearer_token, str):
            raise ApiAuthorizationError("API token passed is not a string.")
        self.bearer_token = bearer_token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Adds the X-Dataverse-key header with the API token and yields the
        original :class:`httpx.Request`.

        Parameters
        ----------
        request : httpx.Request
            The request object which requires authentication headers

        Yields
        ------
        httpx.Request
            The original request with modified headers
        """
        request.headers["Authorization"] = f"Bearer {self.bearer_token}"
        yield request



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------

import pyDataverse
from datetime import date
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

project = "pyDataverse"
author = "Stefan Kasberger"
author_affiliation = "AUSSDA - The Austrian Social Science Data Archive"
copyright = "{0}, {1}".format(date.today().strftime("%Y"), author)
description = "pyDataverse helps with the Dataverse API's and data types (Dataverse, Dataset, Datafile)."

# The short X.Y version
version = pyDataverse.__version__
# The full version, including alpha/beta/rc tags
release = pyDataverse.__version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "code/"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "description": description,
    "show_powered_by": False,
    "github_button": True,
    "github_user": "gdcc",
    "github_repo": "pyDataverse",
    "github_banner": False,
    "travis_button": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
html_sidebars = {
    "index": [
        "sidebar_intro.html",
        "navigation.html",
        "sidebar_related-links.html",
        "sourcelink.html",
        "searchbox.html",
    ],
    "**": [
        "sidebar_intro.html",
        "navigation.html",
        "sidebar_related-links.html",
        "sourcelink.html",
        "searchbox.html",
    ],
}

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "pyDataverse"


# -- Options for LaTeX output ------------------------------------------------

latex_elements: dict[str, str] = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "pyDataverse.tex",
        "pyDataverse Documentation",
        "AUSSDA - Austrian Social Science Data Archive",
        "manual",
    ),
]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "pyDataverse", "pyDataverse Documentation", [author], 1)]

# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "pyDataverse",
        "pyDataverse Documentation",
        author,
        "pyDataverse",
        description,
        "Miscellaneous",
    ),
]


# -- Extension configuration -------------------------------------------------

# Intersphinx

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}



"""Find out more at https://github.com/GDCC/pyDataverse."""


class DataverseError(Exception):
    """Base exception class for Dataverse-related error."""

    pass


class DataverseApiError(DataverseError):
    """Base exception class for Dataverse-related api error."""

    pass


class OperationFailedError(DataverseApiError):
    """Raised when an operation fails for an unknown reason."""

    pass


class ApiUrlError(DataverseApiError):
    """Raised when the request url is not valid."""

    pass


class ApiResponseError(DataverseApiError):
    """Raised when the requests response fails."""

    pass


class ApiAuthorizationError(OperationFailedError):
    """Raised if a user provides invalid credentials."""

    pass


class DataverseNotEmptyError(OperationFailedError):
    """Raised when a Dataverse has accessioned Datasets."""

    pass


class DataverseNotFoundError(OperationFailedError):
    """Raised when a Dataverse cannot be found."""

    pass


class DatasetNotFoundError(OperationFailedError):
    """Raised when a Dataset cannot be found."""

    pass


class DatafileNotFoundError(OperationFailedError):
    """Raised when a Datafile cannot be found."""

    pass



"""Dataverse API wrapper for all it's API's."""

import json
from typing import Any, Dict, Optional
import httpx
import subprocess as sp
from warnings import warn

from httpx import ConnectError, Response

from pyDataverse.auth import ApiTokenAuth
from pyDataverse.exceptions import (
    ApiAuthorizationError,
    ApiUrlError,
    DatasetNotFoundError,
    DataverseNotEmptyError,
    DataverseNotFoundError,
    OperationFailedError,
)

DEPRECATION_GUARD = object()


class Api:
    """Base class.

    Parameters
    ----------
        base_url : str
            Base URL of Dataverse instance. Without trailing `/` at the end.
            e.g. `http://demo.dataverse.org`
        api_token : str
            Authentication token for the api.

    Attributes
    ----------
    base_url
    api_token
    dataverse_version

    """

    def __init__(
        self,
        base_url: str,
        api_token: Optional[str] = None,
        api_version: str = "latest",
        *,
        auth: Optional[httpx.Auth] = None,
    ):
        """Init an Api() class.

        Scheme, host and path combined create the base-url for the api.
        See more about URL at `Wikipedia <https://en.wikipedia.org/wiki/URL>`_.

        Parameters
        ----------
        base_url : str
            Base url for Dataverse api.
        api_token : str | None
            API token for Dataverse API. If you provide an :code:`api_token`, we
            assume it is an API token as retrieved via your Dataverse instance
            user profile.
            We recommend using the :code:`auth` argument instead.
            To retain the current behaviour with the :code:`auth` argument, change

            .. code-block:: python

                Api("https://demo.dataverse.org", "my_token")

            to

            .. code-block:: python

                from pyDataverse.auth import ApiTokenAuth

                Api("https://demo.dataverse.org", auth=ApiTokenAuth("my_token"))

            If you are using an OIDC/OAuth 2.0 Bearer token, please use the :code:`auth`
            parameter with the :py:class:`.auth.BearerTokenAuth`.
        api_version : str
            The version string of the Dataverse API or :code:`latest`, e.g.,
            :code:`v1`. Defaults to :code:`latest`, which drops the version from
            the API urls.
        auth : httpx.Auth | None
            You can provide any authentication mechanism you like to connect to
            your Dataverse instance.  The most common mechanisms are implemented
            in :py:mod:`.auth`, but if one is missing, you can use your own
            `httpx.Auth`-compatible class. For more information, have a look at
            `httpx' Authentication docs
            <https://www.python-httpx.org/advanced/authentication/>`_.
        Examples
        -------
        Create an API connection::

        .. code-block::

            >>> from pyDataverse.api import Api
            >>> base_url = 'http://demo.dataverse.org'
            >>> api = Api(base_url)

        .. code-block::

            >>> from pyDataverse.api import Api
            >>> from pyDataverse.auth import ApiTokenAuth
            >>> base_url = 'http://demo.dataverse.org'
            >>> api = Api(base_url, ApiTokenAuth('my_api_token'))

        """
        if not isinstance(base_url, str):
            raise ApiUrlError("base_url {0} is not a string.".format(base_url))

        self.base_url = base_url.rstrip("/")
        self.client = None

        if not isinstance(api_version, str):
            raise ApiUrlError("api_version {0} is not a string.".format(api_version))
        self.api_version = api_version

        self.auth = auth
        self.api_token = api_token
        if api_token is not None:
            if auth is None:
                self.auth = ApiTokenAuth(api_token)
            else:
                self.api_token = None
                warn(
                    UserWarning(
                        "You provided both, an api_token and a custom auth "
                        "method. We will only use the auth method."
                    )
                )

        if self.api_version == "latest":
            self.base_url_api = "{0}/api".format(self.base_url)
        else:
            self.base_url_api = "{0}/api/{1}".format(self.base_url, self.api_version)

        self.timeout = 500

    def __str__(self):
        """Return the class name and URL of the used API class.

        Returns
        -------
        str
            Naming of the API class.

        """
        return f"{self.__class__.__name__}: {self.base_url_api}"

    def get_request(self, url, params=None, auth=DEPRECATION_GUARD):
        """Make a GET request.

        Parameters
        ----------
        url : str
            Full URL.
        params : dict
            Dictionary of parameters to be passed with the request.
            Defaults to `None`.
        auth : Any
            .. deprecated:: 0.3.4
                The auth parameter was ignored before version 0.3.4.
                Please pass your auth to the Api instance directly, as
                explained in :py:func:`Api.__init__`.
                If you need multiple auth methods, create multiple
                API instances:

                .. code-block:: python

                    api = Api("https://demo.dataverse.org", auth=ApiTokenAuth("my_api_token"))
                    api_oauth = Api("https://demo.dataverse.org", auth=BearerTokenAuth("my_bearer_token"))

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if auth is not DEPRECATION_GUARD:
            warn(
                DeprecationWarning(
                    "The auth parameter is deprecated. Please pass your auth "
                    "arguments to the __init__ method instead."
                )
            )
        headers = {}
        headers["User-Agent"] = "pydataverse"

        if self.client is None:
            return self._sync_request(
                method=httpx.get,
                url=url,
                headers=headers,
                params=params,
            )
        else:
            return self._async_request(
                method=self.client.get,
                url=url,
                headers=headers,
                params=params,
            )

    def post_request(
        self, url, data=None, auth=DEPRECATION_GUARD, params=None, files=None
    ):
        """Make a POST request.

        params will be added as key-value pairs to the URL.

        Parameters
        ----------
        url : str
            Full URL.
        data : str
            Metadata as a json-formatted string. Defaults to `None`.
        auth : Any
            .. deprecated:: 0.3.4
                The auth parameter was ignored before version 0.3.4.
                Please pass your auth to the Api instance directly, as
                explained in :py:func:`Api.__init__`.
                If you need multiple auth methods, create multiple
                API instances:

                .. code-block:: python

                    api = Api("https://demo.dataverse.org", auth=ApiTokenAuth("my_api_token"))
                    api_oauth = Api("https://demo.dataverse.org", auth=BearerTokenAuth("my_bearer_token"))
        files : dict
            e.g. :code:`files={'file': open('sample_file.txt','rb')}`
        params : dict
            Dictionary of parameters to be passed with the request.
            Defaults to :code:`None`.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if auth is not DEPRECATION_GUARD:
            warn(
                DeprecationWarning(
                    "The auth parameter is deprecated. Please pass your auth "
                    "arguments to the __init__ method instead."
                )
            )
        headers = {}
        headers["User-Agent"] = "pydataverse"

        if isinstance(data, str):
            data = json.loads(data)

        # Decide whether to use 'data' or 'json' args
        request_params = self._check_json_data_form(data)

        if self.client is None:
            return self._sync_request(
                method=httpx.post,
                url=url,
                headers=headers,
                params=params,
                files=files,
                **request_params,
            )
        else:
            return self._async_request(
                method=self.client.post,
                url=url,
                headers=headers,
                params=params,
                files=files,
                **request_params,
            )

    def put_request(
        self, url, data=None, auth=DEPRECATION_GUARD, params=None, files=None
    ):
        """Make a PUT request.

        Parameters
        ----------
        url : str
            Full URL.
        data : str
            Metadata as a json-formatted string. Defaults to `None`.
        auth : Any
            .. deprecated:: 0.3.4
                The auth parameter was ignored before version 0.3.4.
                Please pass your auth to the Api instance directly, as
                explained in :py:func:`Api.__init__`.
                If you need multiple auth methods, create multiple
                API instances:

                .. code-block:: python

                    api = Api("https://demo.dataverse.org", auth=ApiTokenAuth("my_api_token"))
                    api_oauth = Api("https://demo.dataverse.org", auth=BearerTokenAuth("my_bearer_token"))
        params : dict
            Dictionary of parameters to be passed with the request.
            Defaults to `None`.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if auth is not DEPRECATION_GUARD:
            warn(
                DeprecationWarning(
                    "The auth parameter is deprecated. Please pass your auth "
                    "arguments to the __init__ method instead."
                )
            )
        headers = {}
        headers["User-Agent"] = "pydataverse"

        if isinstance(data, str):
            data = json.loads(data)

        # Decide whether to use 'data' or 'json' args
        request_params = self._check_json_data_form(data)

        if self.client is None:
            return self._sync_request(
                method=httpx.put,
                url=url,
                headers=headers,
                params=params,
                **request_params,
            )
        else:
            return self._async_request(
                method=self.client.put,
                url=url,
                headers=headers,
                params=params,
                **request_params,
            )

    def delete_request(self, url, auth=DEPRECATION_GUARD, params=None):
        """Make a Delete request.

        Parameters
        ----------
        url : str
            Full URL.
        auth : Any
            .. deprecated:: 0.3.4
                The auth parameter was ignored before version 0.3.4.
                Please pass your auth to the Api instance directly, as
                explained in :py:func:`Api.__init__`.
                If you need multiple auth methods, create multiple
                API instances:

                .. code-block:: python

                    api = Api("https://demo.dataverse.org", auth=ApiTokenAuth("my_api_token"))
                    api_oauth = Api("https://demo.dataverse.org", auth=BearerTokenAuth("my_bearer_token"))
        params : dict
            Dictionary of parameters to be passed with the request.
            Defaults to `None`.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if auth is not DEPRECATION_GUARD:
            warn(
                DeprecationWarning(
                    "The auth parameter is deprecated. Please pass your auth "
                    "arguments to the __init__ method instead."
                )
            )
        headers = {}
        headers["User-Agent"] = "pydataverse"

        if self.client is None:
            return self._sync_request(
                method=httpx.delete,
                url=url,
                headers=headers,
                params=params,
            )
        else:
            return self._async_request(
                method=self.client.delete,
                url=url,
                headers=headers,
                params=params,
            )

    @staticmethod
    def _check_json_data_form(data: Optional[Dict]):
        """This method checks and distributes given payload to match Dataverse expectations.

        In the case of the form-data keyed by "jsonData", Dataverse expects
        the payload as a string in a form of a dictionary. This is not possible
        using HTTPXs json parameter, so we need to handle this case separately.
        """

        if not data:
            return {}
        elif not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        elif "jsonData" not in data:
            return {"json": data}

        assert list(data.keys()) == [
            "jsonData"
        ], "jsonData must be the only key in the dictionary."

        # Content of JSON data should ideally be a string
        content = data["jsonData"]
        if not isinstance(content, str):
            data["jsonData"] = json.dumps(content)

        return {"data": data}

    def _sync_request(
        self,
        method,
        **kwargs,
    ):
        """
        Sends a synchronous request to the specified URL using the specified HTTP method.

        Args:
            method (function): The HTTP method to use for the request.
            **kwargs: Additional keyword arguments to be passed to the method.

        Returns:
            httpx.Response: The response object returned by the request.

        Raises:
            ApiAuthorizationError: If the response status code is 401 (Authorization error).
            ConnectError: If a connection to the API cannot be established.
        """
        assert "url" in kwargs, "URL is required for a request."

        kwargs = self._filter_kwargs(kwargs)

        try:
            resp: httpx.Response = method(
                **kwargs, auth=self.auth, follow_redirects=True, timeout=None
            )
            if resp.status_code == 401:
                try:
                    error_msg = resp.json()["message"]
                except json.JSONDecodeError:
                    error_msg = resp.reason_phrase
                raise ApiAuthorizationError(
                    "ERROR: HTTP 401 - Authorization error {0}. MSG: {1}".format(
                        kwargs["url"], error_msg
                    )
                )

            return resp

        except ConnectError:
            raise ConnectError(
                "ERROR - Could not establish connection to api '{0}'.".format(
                    kwargs["url"]
                )
            )

    async def _async_request(
        self,
        method,
        **kwargs,
    ):
        """
        Sends an asynchronous request to the specified URL using the specified HTTP method.

        Args:
            method (callable): The HTTP method to use for the request.
            **kwargs: Additional keyword arguments to be passed to the method.

        Raises:
            ApiAuthorizationError: If the response status code is 401 (Authorization error).
            ConnectError: If a connection to the API cannot be established.

        Returns:
            The response object.

        """
        assert "url" in kwargs, "URL is required for a request."

        kwargs = self._filter_kwargs(kwargs)

        try:
            resp = await method(**kwargs, auth=self.auth)

            if resp.status_code == 401:
                error_msg = resp.json()["message"]
                raise ApiAuthorizationError(
                    "ERROR: HTTP 401 - Authorization error {0}. MSG: {1}".format(
                        kwargs["url"], error_msg
                    )
                )

            return resp

        except ConnectError:
            raise ConnectError(
                "ERROR - Could not establish connection to api '{0}'.".format(
                    kwargs["url"]
                )
            )

    @staticmethod
    def _filter_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Filters out any keyword arguments that are `None` from the specified dictionary.

        Args:
            kwargs (Dict[str, Any]): The dictionary to filter.

        Returns:
            Dict[str, Any]: The filtered dictionary.
        """
        return {k: v for k, v in kwargs.items() if v is not None}

    async def __aenter__(self):
        """
        Context manager method that initializes an instance of httpx.AsyncClient.

        Returns:
            httpx.AsyncClient: An instance of httpx.AsyncClient.
        """
        self.client = httpx.AsyncClient()

    async def __aexit__(self, exc_type, exc_value, traceback):
        """
        Closes the client connection when exiting a context manager.

        Args:
            exc_type (type): The type of the exception raised, if any.
            exc_value (Exception): The exception raised, if any.
            traceback (traceback): The traceback object associated with the exception, if any.
        """

        await self.client.aclose()
        self.client = None


class DataAccessApi(Api):
    """Class to access Dataverse's Data Access API.

    Attributes
    ----------
    base_url_api_data_access : type
        Description of attribute `base_url_api_data_access`.
    base_url : type
        Description of attribute `base_url`.

    """

    def __init__(self, base_url, api_token=None, *, auth=None):
        """Init an DataAccessApi() class."""
        super().__init__(base_url, api_token, auth=auth)
        if base_url:
            self.base_url_api_data_access = "{0}/access".format(self.base_url_api)
        else:
            self.base_url_api_data_access = self.base_url_api

    def get_datafile(
        self,
        identifier,
        data_format=None,
        no_var_header=None,
        image_thumb=None,
        is_pid=True,
        auth=DEPRECATION_GUARD,
    ):
        """Download a datafile via the Dataverse Data Access API.

        Get by file id (HTTP Request).

        .. code-block:: bash

            GET /api/access/datafile/$id

        Get by persistent identifier (HTTP Request).

        .. code-block:: bash

            GET http://$SERVER/api/access/datafile/:persistentId/?persistentId=doi:10.5072/FK2/J8SJZB

        Parameters
        ----------
        identifier : str
            Identifier of the datafile. Can be datafile id or persistent
            identifier of the datafile (e. g. doi).
        is_pid : bool
            ``True`` to use persistent identifier. ``False``, if not.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        is_first_param = True
        if is_pid:
            url = "{0}/datafile/:persistentId/?persistentId={1}".format(
                self.base_url_api_data_access, identifier
            )
        else:
            url = "{0}/datafile/{1}".format(self.base_url_api_data_access, identifier)
            if data_format or no_var_header or image_thumb:
                url += "?"
        if data_format:
            url += "format={0}".format(data_format)
            is_first_param = False
        if no_var_header:
            if not is_first_param:
                url += "&"
            url += "noVarHeader={0}".format(no_var_header)
            is_first_param = False
        if image_thumb:
            if not is_first_param:
                url += "&"
            url += "imageThumb={0}".format(image_thumb)
        return self.get_request(url, auth=auth)

    def get_datafiles(self, identifier, data_format=None, auth=DEPRECATION_GUARD):
        """Download a datafile via the Dataverse Data Access API.

        Get by file id (HTTP Request).

        .. code-block:: bash

            GET /api/access/datafiles/$id1,$id2,...$idN

        Get by persistent identifier (HTTP Request).

        Parameters
        ----------
        identifier : str
            Identifier of the dataset. Can be datafile id or persistent
            identifier of the datafile (e. g. doi).

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/datafiles/{1}".format(self.base_url_api_data_access, identifier)
        if data_format:
            url += "?format={0}".format(data_format)
        return self.get_request(url, auth=auth)

    def get_datafile_bundle(
        self, identifier, file_metadata_id=None, auth=DEPRECATION_GUARD
    ):
        """Download a datafile in all its formats.

        HTTP Request:

        .. code-block:: bash

            GET /api/access/datafile/bundle/$id

        Data Access API calls can now be made using persistent identifiers (in
        addition to database ids). This is done by passing the constant
        :persistentId where the numeric id of the file is expected, and then
        passing the actual persistent id as a query parameter with the name
        persistentId.

        This is a convenience packaging method available for tabular data
        files. It returns a zipped bundle that contains the data in the
        following formats:
        - Tab-delimited;
        - “Saved Original”, the proprietary (SPSS, Stata, R, etc.) file
        from which the tabular data was ingested;
        - Generated R Data frame (unless the “original” above was in R);
        - Data (Variable) metadata record, in DDI XML;
        - File citation, in Endnote and RIS formats.

        Parameters
        ----------
        identifier : str
            Identifier of the dataset.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/datafile/bundle/{1}".format(
            self.base_url_api_data_access, identifier
        )
        if file_metadata_id:
            url += "?fileMetadataId={0}".format(file_metadata_id)
        return self.get_request(url, auth=auth)

    def request_access(self, identifier, auth=DEPRECATION_GUARD, is_filepid=False):
        """Request datafile access.

        This method requests access to the datafile whose id is passed on the behalf of an authenticated user whose key is passed. Note that not all datasets allow access requests to restricted files.

        https://guides.dataverse.org/en/4.18.1/api/dataaccess.html#request-access

        /api/access/datafile/$id/requestAccess

        curl -H "X-Dataverse-key:$API_TOKEN" -X PUT http://$SERVER/api/access/datafile/{id}/requestAccess
        """
        if is_filepid:
            url = "{0}/datafile/:persistentId/requestAccess?persistentId={1}".format(
                self.base_url_api_data_access, identifier
            )
        else:
            url = "{0}/datafile/{1}/requestAccess".format(
                self.base_url_api_data_access, identifier
            )
        return self.put_request(url, auth=auth)

    def allow_access_request(
        self, identifier, do_allow=True, auth=DEPRECATION_GUARD, is_pid=True
    ):
        """Allow access request for datafiles.

        https://guides.dataverse.org/en/latest/api/dataaccess.html#allow-access-requests

        curl -H "X-Dataverse-key:$API_TOKEN" -X PUT -d true http://$SERVER/api/access/{id}/allowAccessRequest
        curl -H "X-Dataverse-key:$API_TOKEN" -X PUT -d true http://$SERVER/api/access/:persistentId/allowAccessRequest?persistentId={pid}
        """
        if is_pid:
            url = "{0}/:persistentId/allowAccessRequest?persistentId={1}".format(
                self.base_url_api_data_access, identifier
            )
        else:
            url = "{0}/{1}/allowAccessRequest".format(
                self.base_url_api_data_access, identifier
            )

        if do_allow:
            data = "true"
        else:
            data = "false"
        return self.put_request(url, data=data, auth=auth)

    def grant_file_access(self, identifier, user, auth=DEPRECATION_GUARD):
        """Grant datafile access.

        https://guides.dataverse.org/en/4.18.1/api/dataaccess.html#grant-file-access

        curl -H "X-Dataverse-key:$API_TOKEN" -X PUT http://$SERVER/api/access/datafile/{id}/grantAccess/{@userIdentifier}
        """
        url = "{0}/datafile/{1}/grantAccess/{2}".format(
            self.base_url_api_data_access, identifier, user
        )
        return self.put_request(url, auth=auth)

    def list_file_access_requests(self, identifier, auth=DEPRECATION_GUARD):
        """Liste datafile access requests.

        https://guides.dataverse.org/en/4.18.1/api/dataaccess.html#list-file-access-requests

        curl -H "X-Dataverse-key:$API_TOKEN" -X GET http://$SERVER/api/access/datafile/{id}/listRequests
        """
        url = "{0}/datafile/{1}/listRequests".format(
            self.base_url_api_data_access, identifier
        )
        return self.get_request(url, auth=auth)


class MetricsApi(Api):
    """Class to access Dataverse's Metrics API.

    Attributes
    ----------
    base_url_api_metrics : type
        Description of attribute `base_url_api_metrics`.
    base_url : type
        Description of attribute `base_url`.

    """

    def __init__(self, base_url, api_token=None, api_version="latest", *, auth=None):
        """Init an MetricsApi() class."""
        super().__init__(base_url, api_token, api_version, auth=auth)
        if base_url:
            self.base_url_api_metrics = "{0}/api/info/metrics".format(self.base_url)
        else:
            self.base_url_api_metrics = None

    def total(self, data_type, date_str=None, auth=DEPRECATION_GUARD):
        """
        GET https://$SERVER/api/info/metrics/$type
        GET https://$SERVER/api/info/metrics/$type/toMonth/$YYYY-DD

        $type can be set to dataverses, datasets, files or downloads.

        """
        url = "{0}/{1}".format(self.base_url_api_metrics, data_type)
        if date_str:
            url += "/toMonth/{0}".format(date_str)
        return self.get_request(url, auth=auth)

    def past_days(self, data_type, days_str, auth=DEPRECATION_GUARD):
        """

        http://guides.dataverse.org/en/4.18.1/api/metrics.html
        GET https://$SERVER/api/info/metrics/$type/pastDays/$days

        $type can be set to dataverses, datasets, files or downloads.
        """
        # TODO: check if date-string has proper format
        url = "{0}/{1}/pastDays/{2}".format(
            self.base_url_api_metrics, data_type, days_str
        )
        return self.get_request(url, auth=auth)

    def get_dataverses_by_subject(self, auth=DEPRECATION_GUARD):
        """
        GET https://$SERVER/api/info/metrics/dataverses/bySubject

        $type can be set to dataverses, datasets, files or downloads.
        """
        # TODO: check if date-string has proper format
        url = "{0}/dataverses/bySubject".format(self.base_url_api_metrics)
        return self.get_request(url, auth=auth)

    def get_dataverses_by_category(self, auth=DEPRECATION_GUARD):
        """
        GET https://$SERVER/api/info/metrics/dataverses/byCategory

        $type can be set to dataverses, datasets, files or downloads.
        """
        # TODO: check if date-string has proper format
        url = "{0}/dataverses/byCategory".format(self.base_url_api_metrics)
        return self.get_request(url, auth=auth)

    def get_datasets_by_subject(self, date_str=None, auth=DEPRECATION_GUARD):
        """
        GET https://$SERVER/api/info/metrics/datasets/bySubject

        $type can be set to dataverses, datasets, files or downloads.
        """
        # TODO: check if date-string has proper format
        url = "{0}/datasets/bySubject".format(self.base_url_api_metrics)
        if date_str:
            url += "/toMonth/{0}".format(date_str)
        return self.get_request(url, auth=auth)

    def get_datasets_by_data_location(self, data_location, auth=DEPRECATION_GUARD):
        """
        GET https://$SERVER/api/info/metrics/datasets/?dataLocation=$location

        $type can be set to dataverses, datasets, files or downloads.
        """
        # TODO: check if date-string has proper format
        url = "{0}/datasets/?dataLocation={1}".format(
            self.base_url_api_metrics, data_location
        )
        return self.get_request(url, auth=auth)


class NativeApi(Api):
    """Class to access Dataverse's Native API.

    Parameters
    ----------
    base_url : type
        Description of parameter `base_url`.
    api_token : type
        Description of parameter `api_token`.
    api_version : type
        Description of parameter `api_version`.

    Attributes
    ----------
    base_url_api_native : type
        Description of attribute `base_url_api_native`.
    base_url_api : type
        Description of attribute `base_url_api`.

    """

    def __init__(self, base_url: str, api_token=None, api_version="v1", *, auth=None):
        """Init an Api() class.

        Scheme, host and path combined create the base-url for the api.
        See more about URL at `Wikipedia <https://en.wikipedia.org/wiki/URL>`_.

        Parameters
        ----------
        native_api_version : str
            API version of Dataverse native API. Default is `v1`.

        """
        super().__init__(base_url, api_token, api_version, auth=auth)
        self.base_url_api_native = self.base_url_api

    def get_dataverse(self, identifier, auth=DEPRECATION_GUARD):
        """Get dataverse metadata by alias or id.

        View metadata about a dataverse.

        .. code-block:: bash

            GET http://$SERVER/api/dataverses/$id

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long), a dataverse alias (more
            robust), or the special value ``:root``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}".format(self.base_url_api_native, identifier)
        return self.get_request(url, auth=auth)

    def create_dataverse(
        self, parent: str, metadata: str, auth: bool = True
    ) -> Response:
        """Create a dataverse.

        Generates a new dataverse under identifier. Expects a JSON content
        describing the dataverse.

        HTTP Request:

        .. code-block:: bash

            POST http://$SERVER/api/dataverses/$id

        Download the `dataverse.json <http://guides.dataverse.org/en/latest/
        _downloads/dataverse-complete.json>`_ example file and modify to create
        dataverses to suit your needs. The fields name, alias, and
        dataverseContacts are required.

        Status Codes:
            200: dataverse created
            201: dataverse created

        Parameters
        ----------
        parent : str
            Parent dataverse, to which the Dataverse gets attached to.
        metadata : str
            Metadata of the Dataverse.
        auth : bool
            True if api authorization is necessary. Defaults to ``True``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        metadata_dict = json.loads(metadata)
        identifier = metadata_dict["alias"]
        url = "{0}/dataverses/{1}".format(self.base_url_api_native, parent)
        resp = self.post_request(url, metadata, auth)

        if resp.status_code == 404:
            error_msg = resp.json()["message"]
            raise DataverseNotFoundError(
                "ERROR: HTTP 404 - Dataverse {0} was not found. MSG: {1}".format(
                    parent, error_msg
                )
            )
        elif resp.status_code != 200 and resp.status_code != 201:
            error_msg = resp.json()["message"]
            raise OperationFailedError(
                "ERROR: HTTP {0} - Dataverse {1} could not be created. MSG: {2}".format(
                    resp.status_code, identifier, error_msg
                )
            )
        else:
            print("Dataverse {0} created.".format(identifier))
        return resp

    def publish_dataverse(self, identifier, auth=True):
        """Publish a dataverse.

        Publish the Dataverse pointed by identifier, which can either by the
        dataverse alias or its numerical id.

        HTTP Request:

        .. code-block:: bash

            POST http://$SERVER/api/dataverses/$identifier/actions/:publish

        Status Code:
            200: Dataverse published

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long) or a dataverse alias (more
            robust).
        auth : bool
            True if api authorization is necessary. Defaults to ``False``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}/actions/:publish".format(
            self.base_url_api_native, identifier
        )
        resp = self.post_request(url, auth=auth)

        if resp.status_code == 401:
            error_msg = resp.json()["message"]
            raise ApiAuthorizationError(
                "ERROR: HTTP 401 - Publish Dataverse {0} unauthorized. MSG: {1}".format(
                    identifier, error_msg
                )
            )
        elif resp.status_code == 404:
            error_msg = resp.json()["message"]
            raise DataverseNotFoundError(
                "ERROR: HTTP 404 - Dataverse {0} was not found. MSG: {1}".format(
                    identifier, error_msg
                )
            )
        elif resp.status_code != 200:
            error_msg = resp.json()["message"]
            raise OperationFailedError(
                "ERROR: HTTP {0} - Dataverse {1} could not be published. MSG: {2}".format(
                    resp.status_code, identifier, error_msg
                )
            )
        elif resp.status_code == 200:
            print("Dataverse {0} published.".format(identifier))
        return resp

    def delete_dataverse(self, identifier, auth=True):
        """Delete dataverse by alias or id.

        Status Code:
            200: Dataverse deleted

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long) or a dataverse alias (more
            robust).

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}".format(self.base_url_api_native, identifier)
        resp = self.delete_request(url, auth)

        if resp.status_code == 401:
            error_msg = resp.json()["message"]
            raise ApiAuthorizationError(
                "ERROR: HTTP 401 - Delete Dataverse {0} unauthorized. MSG: {1}".format(
                    identifier, error_msg
                )
            )
        elif resp.status_code == 404:
            error_msg = resp.json()["message"]
            raise DataverseNotFoundError(
                "ERROR: HTTP 404 - Dataverse {0} was not found. MSG: {1}".format(
                    identifier, error_msg
                )
            )
        elif resp.status_code == 403:
            error_msg = resp.json()["message"]
            raise DataverseNotEmptyError(
                "ERROR: HTTP 403 - Dataverse {0} not empty. MSG: {1}".format(
                    identifier, error_msg
                )
            )
        elif resp.status_code != 200:
            error_msg = resp.json()["message"]
            raise OperationFailedError(
                "ERROR: HTTP {0} - Dataverse {1} could not be deleted. MSG: {2}".format(
                    resp.status_code, identifier, error_msg
                )
            )
        elif resp.status_code == 200:
            print("Dataverse {0} deleted.".format(identifier))
        return resp

    def get_dataverse_roles(self, identifier: str, auth: bool = False) -> Response:
        """All the roles defined directly in the dataverse by identifier.

        `Docs <https://guides.dataverse.org/en/latest/api/native-api.html#list-roles-defined-in-a-dataverse>`_

        .. code-block:: bash

            GET http://$SERVER/api/dataverses/$id/roles

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long), a dataverse alias (more
            robust), or the special value ``:root``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}/roles".format(self.base_url_api_native, identifier)
        return self.get_request(url, auth=auth)

    def get_dataverse_contents(self, identifier, auth=True):
        """Gets contents of Dataverse.

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long), a dataverse alias (more
            robust), or the special value ``:root``.
        auth : bool
            Description of parameter `auth` (the default is False).

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}/contents".format(self.base_url_api_native, identifier)
        return self.get_request(url, auth=auth)

    def get_dataverse_assignments(self, identifier, auth=DEPRECATION_GUARD):
        """Get dataverse assignments by alias or id.

        View assignments of a dataverse.

        .. code-block:: bash

            GET http://$SERVER/api/dataverses/$id/assignments

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long), a dataverse alias (more
            robust), or the special value ``:root``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}/assignments".format(
            self.base_url_api_native, identifier
        )
        return self.get_request(url, auth=auth)

    def get_dataverse_facets(self, identifier, auth=DEPRECATION_GUARD):
        """Get dataverse facets by alias or id.

        View facets of a dataverse.

        .. code-block:: bash

            GET http://$SERVER/api/dataverses/$id/facets

        Parameters
        ----------
        identifier : str
            Can either be a dataverse id (long), a dataverse alias (more
            robust), or the special value ``:root``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/dataverses/{1}/facets".format(self.base_url_api_native, identifier)
        return self.get_request(url, auth=auth)

    def dataverse_id2alias(self, dataverse_id, auth=DEPRECATION_GUARD):
        """Converts a Dataverse ID to an alias.

        Parameters
        ----------
        dataverse_id : str
            Dataverse ID.

        Returns
        -------
        str
            Dataverse alias

        """
        resp = self.get_dataverse(dataverse_id, auth=auth)
        if "data" in resp.json():
            if "alias" in resp.json()["data"]:
                return resp.json()["data"]["alias"]
        print("ERROR: Can not resolve Dataverse ID to alias.")
        return False

    def get_dataset(self, identifier, version=":latest", auth=True, is_pid=True):
        """Get metadata of a Dataset.

        With Dataverse identifier:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/$identifier

        With persistent identifier:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/:persistentId/?persistentId=$id
            GET http://$SERVER/api/datasets/:persistentId/
            ?persistentId=$pid

        Parameters
        ----------
        identifier : str
            Identifier of the dataset. Can be a Dataverse identifier or a
            persistent identifier (e.g. ``doi:10.11587/8H3N93``).
        is_pid : bool
            True, if identifier is a persistent identifier.
        version : str
            Version to be retrieved:
            ``:latest-published``: the latest published version
            ``:latest``: either a draft (if exists) or the latest published version.
            ``:draft``: the draft version, if any
            ``x.y``: x.y a specific version, where x is the major version number and y is the minor version number.
            ``x``: same as x.0

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if is_pid:
            # TODO: Add version to query http://guides.dataverse.org/en/4.18.1/api/native-api.html#get-json-representation-of-a-dataset
            url = "{0}/datasets/:persistentId/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}".format(self.base_url_api_native, identifier)
            # CHECK: Its not really clear, if the version query can also be done via ID.
        return self.get_request(url, auth=auth)

    def get_dataset_versions(self, identifier, auth=True, is_pid=True):
        """Get versions of a Dataset.

        With Dataverse identifier:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/$identifier/versions

        With persistent identifier:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/:persistentId/versions?persistentId=$id

        Parameters
        ----------
        identifier : str
            Identifier of the dataset. Can be a Dataverse identifier or a
            persistent identifier (e.g. ``doi:10.11587/8H3N93``).
        is_pid : bool
            True, if identifier is a persistent identifier.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/versions?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}/versions".format(
                self.base_url_api_native, identifier
            )
        return self.get_request(url, auth=auth)

    def get_dataset_version(self, identifier, version, auth=True, is_pid=True):
        """Get version of a Dataset.

        With Dataverse identifier:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/$identifier/versions/$versionNumber

        With persistent identifier:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/:persistentId/versions/$versionNumber?persistentId=$id

        Parameters
        ----------
        identifier : str
            Identifier of the dataset. Can be a Dataverse identifier or a
            persistent identifier (e.g. ``doi:10.11587/8H3N93``).
        version : str
            Version string of the Dataset.
        is_pid : bool
            True, if identifier is a persistent identifier.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/versions/{1}?persistentId={2}".format(
                self.base_url_api_native, version, identifier
            )
        else:
            url = "{0}/datasets/{1}/versions/{2}".format(
                self.base_url_api_native, identifier, version
            )
        return self.get_request(url, auth=auth)

    def get_dataset_export(self, pid, export_format, auth=DEPRECATION_GUARD):
        """Get metadata of dataset exported in different formats.

        Export the metadata of the current published version of a dataset
        in various formats by its persistend identifier.

        .. code-block:: bash

            GET http://$SERVER/api/datasets/export?exporter=$exportformat&persistentId=$pid

        Parameters
        ----------
        pid : str
            Persistent identifier of the dataset. (e.g. ``doi:10.11587/8H3N93``).
        export_format : str
            Export format as a string. Formats: ``ddi``, ``oai_ddi``,
            ``dcterms``, ``oai_dc``, ``schema.org``, ``dataverse_json``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/datasets/export?exporter={1}&persistentId={2}".format(
            self.base_url_api_native, export_format, pid
        )
        return self.get_request(url, auth=auth)

    def create_dataset(self, dataverse, metadata, pid=None, publish=False, auth=True):
        """Add dataset to a dataverse.

        `Dataverse Documentation
        <http://guides.dataverse.org/en/latest/api/native-api.html#create-a-dataset-in-a-dataverse>`_

        HTTP Request:

        .. code-block:: bash

            POST http://$SERVER/api/dataverses/$dataverse/datasets --upload-file FILENAME

        Add new dataset with curl:

        .. code-block:: bash

            curl -H "X-Dataverse-key: $API_TOKEN" -X POST $SERVER_URL/api/dataverses/$DV_ALIAS/datasets --upload-file tests/data/dataset_min.json

        Import dataset with existing persistend identifier with curl:

        .. code-block:: bash

            curl -H "X-Dataverse-key: $API_TOKEN" -X POST $SERVER_URL/api/dataverses/$DV_ALIAS/datasets/:import?pid=$PERSISTENT_IDENTIFIER&release=yes --upload-file tests/data/dataset_min.json

        To create a dataset, you must create a JSON file containing all the
        metadata you want such as example file: `dataset-finch1.json
        <http://guides.dataverse.org/en/latest/_downloads/dataset-finch1.json>`_.
        Then, you must decide which dataverse to create the dataset in and
        target that datavese with either the "alias" of the dataverse (e.g.
        "root") or the database id of the dataverse (e.g. "1"). The initial
        version state will be set to "DRAFT":

        Status Code:
            201: dataset created

        Import Dataset with existing PID:
        `<http://guides.dataverse.org/en/latest/api/native-api.html#import-a-dataset-into-a-dataverse>`_
        To import a dataset with an existing persistent identifier (PID), the
        dataset’s metadata should be prepared in Dataverse’s native JSON format. The
        PID is provided as a parameter at the URL. The following line imports a
        dataset with the PID PERSISTENT_IDENTIFIER to Dataverse, and then releases it:

        The pid parameter holds a persistent identifier (such as a DOI or Handle). The import will fail if no PID is provided, or if the provided PID fails validation.

        The optional release parameter tells Dataverse to immediately publish the
        dataset. If the parameter is changed to no, the imported dataset will
        remain in DRAFT status.

        Parameters
        ----------
        dataverse : str
            "alias" of the dataverse (e.g. ``root``) or the database id of the
            dataverse (e.g. ``1``)
        pid : str
            PID of existing Dataset.
        publish : bool
            Publish only works when a Dataset with an existing PID is created. If it
            is ``True``, Dataset should be instantly published, ``False``
            if a Draft should be created.
        metadata : str
            Metadata of the Dataset as a json-formatted string (e. g.
            `dataset-finch1.json <http://guides.dataverse.org/en/latest/_downloads/dataset-finch1.json>`_)

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if pid:
            assert isinstance(pid, str)
            url = "{0}/dataverses/{1}/datasets/:import?pid={2}".format(
                self.base_url_api_native, dataverse, pid
            )
            if publish:
                url += "&release=yes"
            else:
                url += "&release=no"
        else:
            url = "{0}/dataverses/{1}/datasets".format(
                self.base_url_api_native, dataverse
            )
        resp = self.post_request(url, metadata, auth)

        if resp.status_code == 404:
            error_msg = resp.json()["message"]
            raise DataverseNotFoundError(
                "ERROR: HTTP 404 - Dataverse {0} was not found. MSG: {1}".format(
                    dataverse, error_msg
                )
            )
        elif resp.status_code == 401:
            error_msg = resp.json()["message"]
            raise ApiAuthorizationError(
                "ERROR: HTTP 401 - Create Dataset unauthorized. MSG: {0}".format(
                    error_msg
                )
            )
        elif resp.status_code == 201:
            if "data" in resp.json():
                if "persistentId" in resp.json()["data"]:
                    identifier = resp.json()["data"]["persistentId"]
                    print("Dataset with pid '{0}' created.".format(identifier))
                elif "id" in resp.json()["data"]:
                    identifier = resp.json()["data"]["id"]
                    print("Dataset with id '{0}' created.".format(identifier))
                else:
                    print("ERROR: No identifier returned for created Dataset.")
        return resp

    def edit_dataset_metadata(
        self, identifier, metadata, is_pid=True, replace=False, auth=True
    ):
        """Edit metadata of a given dataset.

        `edit-dataset-metadata <http://guides.dataverse.org/en/latest/api/native-api.html#edit-dataset-metadata>`_.

        HTTP Request:

        .. code-block:: bash

            PUT http://$SERVER/api/datasets/editMetadata/$id --upload-file FILENAME

        Add data to dataset fields that are blank or accept multiple values with
        the following

        CURL Request:

        .. code-block:: bash

            curl -H "X-Dataverse-key: $API_TOKEN" -X PUT $SERVER_URL/api/datasets/:persistentId/editMetadata/?persistentId=$pid --upload-file dataset-add-metadata.json

        For these edits your JSON file need only include those dataset fields
        which you would like to edit. A sample JSON file may be downloaded
        here: `dataset-edit-metadata-sample.json
        <http://guides.dataverse.org/en/latest/_downloads/dataset-finch1.json>`_

        Parameters
        ----------
        identifier : str
            Identifier of the dataset. Can be a Dataverse identifier or a
            persistent identifier (e.g. ``doi:10.11587/8H3N93``).
        metadata : str
            Metadata of the Dataset as a json-formatted string.
        is_pid : bool
            ``True`` to use persistent identifier. ``False``, if not.
        replace : bool
            ``True`` to replace already existing metadata. ``False``, if not.
        auth : bool
            ``True``, if an api token should be sent. Defaults to ``False``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        Examples
        -------
        Get dataset metadata::

            >>> data = api.get_dataset(doi).json()["data"]["latestVersion"]["metadataBlocks"]["citation"]
            >>> resp = api.edit_dataset_metadata(doi, data, is_replace=True, auth=True)
            >>> resp.status_code
            200: metadata updated

        """
        params = {"replace": "true"} if replace else {}

        if is_pid:
            url = "{0}/datasets/:persistentId/editMetadata".format(
                self.base_url_api_native
            )
            params["persistentId"] = identifier
        else:
            url = "{0}/datasets/editMetadata/{1}".format(
                self.base_url_api_native, identifier
            )

        resp = self.put_request(
            url=url,
            data=metadata,
            auth=auth,
            params=params,
        )

        if resp.status_code == 401:
            error_msg = resp.json()["message"]
            raise ApiAuthorizationError(
                "ERROR: HTTP 401 - Updating metadata unauthorized. MSG: {0}".format(
                    error_msg
                )
            )
        elif resp.status_code == 400:
            if "Error parsing" in resp.json()["message"]:
                print("Wrong passed data format.")
            else:
                print(
                    "You may not add data to a field that already has data and does not"
                    " allow multiples. Use is_replace=true to replace existing data."
                )
        elif resp.status_code == 200:
            print("Dataset '{0}' updated".format(identifier))
        return resp

    def create_dataset_private_url(self, identifier, is_pid=True, auth=True):
        """Create private Dataset URL.

        POST http://$SERVER/api/datasets/$id/privateUrl?key=$apiKey


        http://guides.dataverse.org/en/4.16/api/native-api.html#create-a-private-url-for-a-dataset
                'MSG: {1}'.format(pid, error_msg))

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/privateUrl/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}/privateUrl".format(
                self.base_url_api_native, identifier
            )

        resp = self.post_request(url, auth=auth)

        if resp.status_code == 200:
            print(
                "Dataset private URL created: {0}".format(resp.json()["data"]["link"])
            )
        return resp

    def get_dataset_private_url(self, identifier, is_pid=True, auth=True):
        """Get private Dataset URL.

        GET http://$SERVER/api/datasets/$id/privateUrl?key=$apiKey

        http://guides.dataverse.org/en/4.16/api/native-api.html#get-the-private-url-for-a-dataset

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/privateUrl/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}/privateUrl".format(
                self.base_url_api_native, identifier
            )

        resp = self.get_request(url, auth=auth)

        if resp.status_code == 200:
            print("Got Dataset private URL: {0}".format(resp.json()["data"]["link"]))
        return resp

    def delete_dataset_private_url(self, identifier, is_pid=True, auth=True):
        """Get private Dataset URL.

        DELETE http://$SERVER/api/datasets/$id/privateUrl?key=$apiKey

        http://guides.dataverse.org/en/4.16/api/native-api.html#delete-the-private-url-from-a-dataset

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/privateUrl/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}/privateUrl".format(
                self.base_url_api_native, identifier
            )

        resp = self.delete_request(url, auth=auth)

        if resp.status_code == 200:
            print("Got Dataset private URL: {0}".format(resp.json()["data"]["link"]))
        return resp

    def publish_dataset(self, pid, release_type="minor", auth=True):
        """Publish dataset.

        Publishes the dataset whose id is passed. If this is the first version
        of the dataset, its version number will be set to 1.0. Otherwise, the
        new dataset version number is determined by the most recent version
        number and the type parameter. Passing type=minor increases the minor
        version number (2.3 is updated to 2.4). Passing type=major increases
        the major version number (2.3 is updated to 3.0). Superusers can pass
        type=updatecurrent to update metadata without changing the version
        number.

        HTTP Request:

        .. code-block:: bash

            POST http://$SERVER/api/datasets/$id/actions/:publish?type=$type

        When there are no default workflows, a successful publication process
        will result in 200 OK response. When there are workflows, it is
        impossible for Dataverse to know how long they are going to take and
        whether they will succeed or not (recall that some stages might require
        human intervention). Thus, a 202 ACCEPTED is returned immediately. To
        know whether the publication process succeeded or not, the client code
        has to check the status of the dataset periodically, or perform some
        push request in the post-publish workflow.

        Status Code:
            200: dataset published

        Parameters
        ----------
        pid : str
            Persistent identifier of the dataset (e.g.
            ``doi:10.11587/8H3N93``).
        release_type : str
            Passing ``minor`` increases the minor version number (2.3 is
            updated to 2.4). Passing ``major`` increases the major version
            number (2.3 is updated to 3.0). Superusers can pass
            ``updatecurrent`` to update metadata without changing the version
            number.
        auth : bool
            ``True`` if api authorization is necessary. Defaults to ``False``.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/datasets/:persistentId/actions/:publish".format(
            self.base_url_api_native
        )
        url += "?persistentId={0}&type={1}".format(pid, release_type)
        resp = self.post_request(url, auth=auth)

        if resp.status_code == 404:
            error_msg = resp.json()["message"]
            raise DatasetNotFoundError(
                "ERROR: HTTP 404 - Dataset {0} was not found. MSG: {1}".format(
                    pid, error_msg
                )
            )
        elif resp.status_code == 401:
            error_msg = resp.json()["message"]
            raise ApiAuthorizationError(
                "ERROR: HTTP 401 - User not allowed to publish dataset {0}. "
                "MSG: {1}".format(pid, error_msg)
            )
        elif resp.status_code == 200:
            print("Dataset {0} published".format(pid))
        return resp

    def get_dataset_lock(self, pid):
        """Get if dataset is locked.

        The lock API endpoint was introduced in Dataverse 4.9.3.

        Parameters
        ----------
        pid : str
            Persistent identifier of the Dataset (e.g.
            ``doi:10.11587/8H3N93``).

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/datasets/:persistentId/locks/?persistentId={1}".format(
            self.base_url_api_native, pid
        )
        return self.get_request(url, auth=True)

    def get_dataset_assignments(self, identifier, is_pid=True, auth=True):
        """Get Dataset assignments.

        GET http://$SERVER/api/datasets/$id/assignments?key=$apiKey


        """
        if is_pid:
            url = "{0}/datasets/:persistentId/assignments/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}/assignments".format(
                self.base_url_api_native, identifier
            )
        return self.get_request(url, auth=auth)

    def delete_dataset(self, identifier, is_pid=True, auth=True):
        """Delete a dataset.

        Delete the dataset whose id is passed

        Status Code:
            200: dataset deleted

        Parameters
        ----------
        identifier : str
            Identifier of the dataset. Can be a Dataverse identifier or a
            persistent identifier (e.g. ``doi:10.11587/8H3N93``).
        is_pid : bool
            True, if identifier is a persistent identifier.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}".format(self.base_url_api_native, identifier)
        resp = self.delete_request(url, auth=auth)

        if resp.status_code == 404:
            error_msg = resp.json()["message"]
            raise DatasetNotFoundError(
                "ERROR: HTTP 404 - Dataset '{0}' was not found. MSG: {1}".format(
                    identifier, error_msg
                )
            )
        elif resp.status_code == 405:
            error_msg = resp.json()["message"]
            raise OperationFailedError(
                "ERROR: HTTP 405 - "
                "Published datasets can only be deleted from the GUI. For "
                "more information, please refer to "
                "https://github.com/IQSS/dataverse/issues/778"
                " MSG: {0}".format(error_msg)
            )
        elif resp.status_code == 401:
            error_msg = resp.json()["message"]
            raise ApiAuthorizationError(
                "ERROR: HTTP 401 - User not allowed to delete dataset '{0}'. "
                "MSG: {1}".format(identifier, error_msg)
            )
        elif resp.status_code == 200:
            print("Dataset '{0}' deleted.".format(identifier))
        return resp

    def destroy_dataset(self, identifier, is_pid=True, auth=True):
        """Destroy Dataset.

        http://guides.dataverse.org/en/4.16/api/native-api.html#delete-published-dataset

        Normally published datasets should not be deleted, but there exists a
        “destroy” API endpoint for superusers which will act on a dataset given
        a persistent ID or dataset database ID:

        curl -H "X-Dataverse-key:$API_TOKEN" -X DELETE http://$SERVER/api/datasets/:persistentId/destroy/?persistentId=doi:10.5072/FK2/AAA000

        curl -H "X-Dataverse-key:$API_TOKEN" -X DELETE http://$SERVER/api/datasets/999/destroy

        Calling the destroy endpoint is permanent and irreversible. It will
        remove the dataset and its datafiles, then re-index the parent
        dataverse in Solr. This endpoint requires the API token of a
        superuser.

        """
        if is_pid:
            url = "{0}/datasets/:persistentId/destroy/?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            url = "{0}/datasets/{1}/destroy".format(
                self.base_url_api_native, identifier
            )

        resp = self.delete_request(url, auth=auth)

        if resp.status_code == 200:
            print("Dataset {0} destroyed".format(resp.json()))
        return resp

    def get_datafiles_metadata(self, pid, version=":latest", auth=True):
        """List metadata of all datafiles of a dataset.

        `Documentation <http://guides.dataverse.org/en/latest/api/native-api.html#list-files-in-a-dataset>`_

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/datasets/$id/versions/$versionId/files

        Parameters
        ----------
        pid : str
            Persistent identifier of the dataset. e.g. ``doi:10.11587/8H3N93``.
        version : str
            Version of dataset. Defaults to `1`.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        base_str = "{0}/datasets/:persistentId/versions/".format(
            self.base_url_api_native
        )
        url = base_str + "{0}/files?persistentId={1}".format(version, pid)
        return self.get_request(url, auth=auth)

    def get_datafile_metadata(
        self, identifier, is_filepid=False, is_draft=False, auth=True
    ):
        """
        GET http://$SERVER/api/files/{id}/metadata

        curl $SERVER_URL/api/files/$ID/metadata
        curl "$SERVER_URL/api/files/:persistentId/metadata?persistentId=$PERSISTENT_ID"
        curl "https://demo.dataverse.org/api/files/:persistentId/metadata?persistentId=doi:10.5072/FK2/AAA000"
        curl -H "X-Dataverse-key:$API_TOKEN" $SERVER_URL/api/files/$ID/metadata/draft

        """
        if is_filepid:
            url = "{0}/files/:persistentId/metadata".format(self.base_url_api_native)
            if is_draft:
                url += "/draft"
            url += "?persistentId={0}".format(identifier)
        else:
            url = "{0}/files/{1}/metadata".format(self.base_url_api_native, identifier)
            if is_draft:
                url += "/draft"
            # CHECK: Its not really clear, if the version query can also be done via ID.
        return self.get_request(url, auth=auth)

    def upload_datafile(self, identifier, filename, json_str=None, is_pid=True):
        """Add file to a dataset.

        Add a file to an existing Dataset. Description and tags are optional:

        HTTP Request:

        .. code-block:: bash

            POST http://$SERVER/api/datasets/$id/add

        The upload endpoint checks the content of the file, compares it with
        existing files and tells if already in the database (most likely via
        hashing).

        `adding-files <http://guides.dataverse.org/en/latest/api/native-api.html#adding-files>`_.

        Parameters
        ----------
        identifier : str
            Identifier of the dataset.
        filename : str
            Full filename with path.
        json_str : str
            Metadata as JSON string.
        is_pid : bool
            ``True`` to use persistent identifier. ``False``, if not.

        Returns
        -------
        dict
            The json string responded by the CURL request, converted to a
            dict().

        """
        url = self.base_url_api_native
        if is_pid:
            url += "/datasets/:persistentId/add?persistentId={0}".format(identifier)
        else:
            url += "/datasets/{0}/add".format(identifier)

        files = {"file": open(filename, "rb")}
        metadata = {}
        if json_str is not None:
            metadata["jsonData"] = json_str
        return self.post_request(url, data=metadata, files=files, auth=True)

    def update_datafile_metadata(self, identifier, json_str=None, is_filepid=False):
        """Update datafile metadata.

        metadata such as description, directoryLabel (File Path) and tags are not carried over from the file being replaced:
        Updates the file metadata for an existing file where ID is the
        database id of the file to update or PERSISTENT_ID is the persistent id
        (DOI or Handle) of the file. Requires a jsonString expressing the new
        metadata. No metadata from the previous version of this file will be
        persisted, so if you want to update a specific field first get the
        json with the above command and alter the fields you want.


        Also note that dataFileTags are not versioned and changes to these will update the published version of the file.

        This functions needs CURL to work!

        HTTP Request:

        .. code-block:: bash

            POST -F 'file=@file.extension' -F 'jsonData={json}' http://$SERVER/api/files/{id}/metadata?key={apiKey}
            curl -H "X-Dataverse-key:$API_TOKEN" -X POST -F 'jsonData={"description":"My description bbb.","provFreeform":"Test prov freeform","categories":["Data"],"restrict":false}' $SERVER_URL/api/files/$ID/metadata
            curl -H "X-Dataverse-key:xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx" -X POST -F 'jsonData={"description":"My description bbb.","provFreeform":"Test prov freeform","categories":["Data"],"restrict":false}' "https://demo.dataverse.org/api/files/:persistentId/metadata?persistentId=doi:10.5072/FK2/AAA000"

        `Docs <http://guides.dataverse.org/en/latest/api/native-api.html#updating-file-metadata>`_.

        Parameters
        ----------
        identifier : str
            Identifier of the dataset.
        json_str : str
            Metadata as JSON string.
        is_filepid : bool
            ``True`` to use persistent identifier for datafile. ``False``, if
            not.

        Returns
        -------
        dict
            The json string responded by the CURL request, converted to a
            dict().

        """
        # if is_filepid:
        #     url = '{0}/files/:persistentId/metadata?persistentId={1}'.format(
        #         self.base_url_api_native, identifier)
        # else:
        #     url = '{0}/files/{1}/metadata'.format(self.base_url_api_native, identifier)
        #
        # data = {'jsonData': json_str}
        # resp = self.post_request(
        #     url,
        #     data=data,
        #     auth=True
        #     )
        query_str = self.base_url_api_native
        if is_filepid:
            query_str = "{0}/files/:persistentId/metadata?persistentId={1}".format(
                self.base_url_api_native, identifier
            )
        else:
            query_str = "{0}/files/{1}/metadata".format(
                self.base_url_api_native, identifier
            )
        shell_command = 'curl -H "X-Dataverse-key: {0}"'.format(self.api_token)
        shell_command += " -X POST -F 'jsonData={0}' {1}".format(json_str, query_str)
        # TODO(Shell): is shell=True necessary?
        return sp.run(shell_command, shell=True, stdout=sp.PIPE)

    def replace_datafile(self, identifier, filename, json_str, is_filepid=True):
        """Replace datafile.

        HTTP Request:

        .. code-block:: bash

            POST -F 'file=@file.extension' -F 'jsonData={json}' http://$SERVER/api/files/{id}/replace?key={apiKey}

        `replacing-files <http://guides.dataverse.org/en/latest/api/native-api.html#replacing-files>`_.

        Parameters
        ----------
        identifier : str
            Identifier of the file to be replaced.
        filename : str
            Full filename with path.
        json_str : str
            Metadata as JSON string.
        is_filepid : bool
            ``True`` if ``identifier`` is a persistent identifier for the datafile.
            ``False``, if not.

        Returns
        -------
        dict
            The json string responded by the CURL request, converted to a
            dict().

        """
        url = self.base_url_api_native
        files = {"file": open(filename, "rb")}
        data = {"jsonData": json_str}

        if is_filepid:
            url += "/files/:persistentId/replace?persistentId={0}".format(identifier)
        else:
            url += "/files/{0}/replace".format(identifier)
        return self.post_request(url, data=data, files=files, auth=True)

    def get_info_version(self, auth=DEPRECATION_GUARD):
        """Get the Dataverse version and build number.

        The response contains the version and build numbers. Requires no api
        token.

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/info/version

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/info/version".format(self.base_url_api_native)
        return self.get_request(url, auth=auth)

    def get_info_server(self, auth=DEPRECATION_GUARD):
        """Get dataverse server name.

        This is useful when a Dataverse system is composed of multiple Java EE
        servers behind a load balancer.

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/info/server

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/info/server".format(self.base_url_api_native)
        return self.get_request(url, auth=auth)

    def get_info_api_terms_of_use(self, auth=DEPRECATION_GUARD):
        """Get API Terms of Use url.

        The response contains the text value inserted as API Terms of use which
        uses the database setting :ApiTermsOfUse.

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/info/apiTermsOfUse

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/info/apiTermsOfUse".format(self.base_url_api_native)
        return self.get_request(url, auth=auth)

    def get_metadatablocks(self, auth=DEPRECATION_GUARD):
        """Get info about all metadata blocks.

        Lists brief info about all metadata blocks registered in the system.

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/metadatablocks

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/metadatablocks".format(self.base_url_api_native)
        return self.get_request(url, auth=auth)

    def get_metadatablock(self, identifier, auth=DEPRECATION_GUARD):
        """Get info about single metadata block.

        Returns data about the block whose identifier is passed. identifier can
        either be the block’s id, or its name.

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/metadatablocks/$identifier

        Parameters
        ----------
        identifier : str
            Can be block's id, or it's name.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/metadatablocks/{1}".format(self.base_url_api_native, identifier)
        return self.get_request(url, auth=auth)

    def get_user_api_token_expiration_date(self, auth=DEPRECATION_GUARD):
        """Get the expiration date of an Users's API token.

        HTTP Request:

        .. code-block:: bash

            curl -H X-Dataverse-key:$API_TOKEN -X GET $SERVER_URL/api/users/token

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/users/token".format(self.base_url_api_native)
        return self.get_request(url, auth=auth)

    def recreate_user_api_token(self):
        """Recreate an Users API token.

        HTTP Request:

        .. code-block:: bash

            curl -H X-Dataverse-key:$API_TOKEN -X POST $SERVER_URL/api/users/token/recreate

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/users/token/recreate".format(self.base_url_api_native)
        return self.post_request(url)

    def delete_user_api_token(self):
        """Delete an Users API token.

        HTTP Request:

        .. code-block:: bash

            curl -H X-Dataverse-key:$API_TOKEN -X POST $SERVER_URL/api/users/token/recreate

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/users/token".format(self.base_url_api_native)
        return self.delete_request(url)

    def create_role(self, dataverse_id):
        """Create a new role.

        `Docs <https://guides.dataverse.org/en/latest/api/native-api.html#id2>`_

        HTTP Request:

        .. code-block:: bash

            POST http://$SERVER/api/roles?dvo=$dataverseIdtf&key=$apiKey

        Parameters
        ----------
        dataverse_id : str
            Can be alias or id of a Dataverse.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/roles?dvo={1}".format(self.base_url_api_native, dataverse_id)
        return self.post_request(url)

    def show_role(self, role_id, auth=DEPRECATION_GUARD):
        """Show role.

        `Docs <https://guides.dataverse.org/en/latest/api/native-api.html#show-role>`_

        HTTP Request:

        .. code-block:: bash

            GET http://$SERVER/api/roles/$id

        Parameters
        ----------
        identifier : str
            Can be alias or id of a Dataverse.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/roles/{1}".format(self.base_url_api_native, role_id)
        return self.get_request(url, auth=auth)

    def delete_role(self, role_id):
        """Delete role.

        `Docs <https://guides.dataverse.org/en/latest/api/native-api.html#delete-role>`_

        Parameters
        ----------
        identifier : str
            Can be alias or id of a Dataverse.

        Returns
        -------
        httpx.Response
            Response object of httpx library.

        """
        url = "{0}/roles/{1}".format(self.base_url_api_native, role_id)
        return self.delete_request(url)

    def get_children(
        self, parent=":root", parent_type="dataverse", children_types=None, auth=True
    ):
        """Walk through children of parent element in Dataverse tree.

        Default: gets all child dataverses if parent = dataverse or all

        Example Dataverse Tree:

        .. code-block:: bash

            data = {
                'type': 'dataverse',
                'dataverse_id': 1,
                'dataverse_alias': ':root',
                'children': [
                    {
                        'type': 'datasets',
                        'dataset_id': 231,
                        'pid': 'doi:10.11587/LYFDYC',
                        'children': [
                            {
                                'type': 'datafile'
                                'datafile_id': 532,
                                'pid': 'doi:10.11587/LYFDYC/C2WTRN',
                                'filename': '10082_curation.pdf '
                            }
                        ]
                    }
                ]
            }

        Parameters
        ----------
        parent : str
            Description of parameter `parent`.
        parent_type : str
            Description of parameter `parent_type`.
        children_types : list
            Types of children to be collected. 'dataverses', 'datasets' and 'datafiles' are valid list items.
        auth : bool
            Authentication needed

        Returns
        -------
        list
            List of Dataverse data type dictionaries. Different ones for
            Dataverses, Datasets and Datafiles.

        # TODO
        - differentiate between published and unpublished data types
        - util function to read out all dataverses into a list
        - util function to read out all datasets into a list
        - util function to read out all datafiles into a list
        - Unify tree and models

        """
        children = []

        if children_types is None:
            children_types = []

        if len(children_types) == 0:
            if parent_type == "dataverse":
                children_types = ["dataverses"]
            elif parent_type == "dataset":
                children_types = ["datafiles"]

        if (
            "dataverses" in children_types
            and "datafiles" in children_types
            and "datasets" not in children_types
        ):
            print(
                "ERROR: Wrong children_types passed: 'dataverses' and 'datafiles'"
                " passed, 'datasets' missing."
            )
            return False

        if parent_type == "dataverse":
            # check for dataverses and datasets as children and get their ID
            parent_alias = parent
            resp = self.get_dataverse_contents(parent_alias, auth=auth)
            if "data" in resp.json():
                contents = resp.json()["data"]
                for content in contents:
                    if (
                        content["type"] == "dataverse"
                        and "dataverses" in children_types
                    ):
                        dataverse_id = content["id"]
                        child_alias = self.dataverse_id2alias(dataverse_id, auth=auth)
                        children.append(
                            {
                                "dataverse_id": dataverse_id,
                                "title": content["title"],
                                "dataverse_alias": child_alias,
                                "type": "dataverse",
                                "children": self.get_children(
                                    parent=child_alias,
                                    parent_type="dataverse",
                                    children_types=children_types,
                                    auth=auth,
                                ),
                            }
                        )
                    elif content["type"] == "dataset" and "datasets" in children_types:
                        pid = (
                            content["protocol"]
                            + ":"
                            + content["authority"]
                            + "/"
                            + content["identifier"]
                        )
                        children.append(
                            {
                                "dataset_id": content["id"],
                                "pid": pid,
                                "type": "dataset",
                                "children": self.get_children(
                                    parent=pid,
                                    parent_type="dataset",
                                    children_types=children_types,
                                    auth=auth,
                                ),
                            }
                        )
            else:
                print("ERROR: 'get_dataverse_contents()' API request not working.")
        elif parent_type == "dataset" and "datafiles" in children_types:
            # check for datafiles as children and get their ID
            pid = parent
            resp = self.get_datafiles_metadata(parent, version=":latest")
            if "data" in resp.json():
                for datafile in resp.json()["data"]:
                    children.append(
                        {
                            "datafile_id": datafile["dataFile"]["id"],
                            "filename": datafile["dataFile"]["filename"],
                            "label": datafile["label"],
                            "pid": datafile["dataFile"]["persistentId"],
                            "type": "datafile",
                        }
                    )
            else:
                print("ERROR: 'get_datafiles()' API request not working.")
        return children

    def get_user(self):
        """Get details of the current authenticated user.

        Auth must be ``true`` for this to work. API endpoint is available for Dataverse >= 5.3.

        https://guides.dataverse.org/en/latest/api/native-api.html#get-user-information-in-json-format
        """
        url = f"{self.base_url}/users/:me"
        return self.get_request(url, auth=True)

    def redetect_file_type(
        self, identifier: str, is_pid: bool = False, dry_run: bool = False
    ) -> Response:
        """Redetect file type.

        https://guides.dataverse.org/en/latest/api/native-api.html#redetect-file-type

        Parameters
        ----------
        identifier : str
            Datafile id (fileid) or file PID.
        is_pid : bool
            Is the identifier a PID, by default False.
        dry_run : bool, optional
            [description], by default False

        Returns
        -------
        Response
            Request Response() object.
        """
        if dry_run is True:
            dry_run_str = "true"
        elif dry_run is False:
            dry_run_str = "false"
        if is_pid:
            url = f"{self.base_url_api_native}/files/:persistentId/redetect?persistentId={identifier}&dryRun={dry_run_str}"
        else:
            url = f"{self.base_url_api_native}/files/{identifier}/redetect?dryRun={dry_run_str}"
        return self.post_request(url, auth=True)

    def reingest_datafile(self, identifier: str, is_pid: bool = False) -> Response:
        """Reingest datafile.

        https://guides.dataverse.org/en/latest/api/native-api.html#reingest-a-file

        Parameters
        ----------
        identifier : str
            Datafile id (fileid) or file PID.
        is_pid : bool
            Is the identifier a PID, by default False.

        Returns
        -------
        Response
            Request Response() object.
        """
        if is_pid:
            url = f"{self.base_url_api_native}/files/:persistentId/reingest?persistentId={identifier}"
        else:
            url = f"{self.base_url_api_native}/files/{identifier}/reingest"
        return self.post_request(url, auth=True)

    def uningest_datafile(self, identifier: str, is_pid: bool = False) -> Response:
        """Uningest datafile.

        https://guides.dataverse.org/en/latest/api/native-api.html#uningest-a-file

        Parameters
        ----------
        identifier : str
            Datafile id (fileid) or file PID.
        is_pid : bool
            Is the identifier a PID, by default False.

        Returns
        -------
        Response
            Request Response() object.
        """
        if is_pid:
            url = f"{self.base_url_api_native}/files/:persistentId/uningest?persistentId={identifier}"
        else:
            url = f"{self.base_url_api_native}/files/{identifier}/uningest"
        return self.post_request(url, auth=True)

    def restrict_datafile(self, identifier: str, is_pid: bool = False) -> Response:
        """Uningest datafile.

        https://guides.dataverse.org/en/latest/api/native-api.html#restrict-files

        Parameters
        ----------
        identifier : str
            Datafile id (fileid) or file PID.
        is_pid : bool
            Is the identifier a PID, by default False.

        Returns
        -------
        Response
            Request Response() object.
        """
        if is_pid:
            url = f"{self.base_url_api_native}/files/:persistentId/restrict?persistentId={identifier}"
        else:
            url = f"{self.base_url_api_native}/files/{identifier}/restrict"
        return self.put_request(url, auth=True)


class SearchApi(Api):
    """Class to access Dataverse's Search API.

    Examples
    -------
    Examples should be written in doctest format, and
    should illustrate how to use the function/class.
    >>>

    Attributes
    ----------
    base_url_api_search : type
        Description of attribute `base_url_api_search`.
    base_url : type
        Description of attribute `base_url`.

    """

    def __init__(self, base_url, api_token=None, api_version="latest", *, auth=None):
        """Init an SearchApi() class."""
        super().__init__(base_url, api_token, api_version, auth=auth)
        if base_url:
            self.base_url_api_search = "{0}/search?q=".format(self.base_url_api)
        else:
            self.base_url_api_search = self.base_url_api

    def search(
        self,
        q_str,
        data_type=None,
        subtree=None,
        sort=None,
        order=None,
        per_page=None,
        start=None,
        show_relevance=None,
        show_facets=None,
        filter_query=None,
        show_entity_ids=None,
        query_entities=None,
        auth=DEPRECATION_GUARD,
    ):
        """Search.

        http://guides.dataverse.org/en/4.18.1/api/search.html
        """
        url = "{0}{1}".format(self.base_url_api_search, q_str)
        if data_type:
            # TODO: pass list of types
            url += "&type={0}".format(data_type)
        if subtree:
            # TODO: pass list of subtrees
            url += "&subtree={0}".format(subtree)
        if sort:
            url += "&sort={0}".format(sort)
        if order:
            url += "&order={0}".format(order)
        if per_page:
            url += "&per_page={0}".format(per_page)
        if start:
            url += "&start={0}".format(start)
        if show_relevance:
            url += "&show_relevance={0}".format(show_relevance)
        if show_facets:
            url += "&show_facets={0}".format(show_facets)
        if filter_query:
            url += "&fq={0}".format(filter_query)
        if show_entity_ids:
            url += "&show_entity_ids={0}".format(show_entity_ids)
        if query_entities:
            url += "&query_entities={0}".format(query_entities)
        return self.get_request(url, auth=auth)


class SwordApi(Api):
    """Class to access Dataverse's SWORD API.

    Parameters
    ----------
    sword_api_version : str
        SWORD API version. Defaults to 'v1.1'.

    Attributes
    ----------
    base_url_api_sword : str
        Description of attribute `base_url_api_sword`.
    base_url : str
        Description of attribute `base_url`.
    native_api_version : str
        Description of attribute `native_api_version`.
    sword_api_version

    """

    def __init__(
        self,
        base_url,
        api_version="v1.1",
        api_token=None,
        sword_api_version="v1.1",
        *,
        auth=None,
    ):
        """Init a :class:`SwordApi <pyDataverse.api.SwordApi>` instance.

        Parameters
        ----------
        sword_api_version : str
            Api version of Dataverse SWORD API.
        api_token : str | None
            An Api token as retrieved from your Dataverse instance.
        auth : httpx.Auth
            Note that the SWORD API uses a different authentication mechanism
            than the native API, in particular it uses `HTTP Basic
            Authentication
            <https://guides.dataverse.org/en/latest/api/sword.html#sword-auth>`_.
            Thus, if you pass an api_token, it will be used as the username in
            the HTTP Basic Authentication. If you pass a custom :py:class:`httpx.Auth`, use
            :py:class:`httpx.BasicAuth` with an empty password:

            .. code-block:: python

                sword_api = Api(
                    "https://demo.dataverse.org", auth=httpx.BasicAuth(username="my_token", password="")
                )

        """
        if auth is None and api_token is not None:
            auth = httpx.BasicAuth(api_token, "")
        super().__init__(base_url, api_token, api_version, auth=auth)
        if not isinstance(sword_api_version, ("".__class__, "".__class__)):
            raise ApiUrlError(
                "sword_api_version {0} is not a string.".format(sword_api_version)
            )
        self.sword_api_version = sword_api_version

        # Test connection.
        if self.base_url and sword_api_version:
            self.base_url_api_sword = "{0}/dvn/api/data-deposit/{1}".format(
                self.base_url, self.sword_api_version
            )
        else:
            self.base_url_api_sword = base_url

    def get_service_document(self):
        url = "{0}/swordv2/service-document".format(self.base_url_api_sword)
        return self.get_request(url, auth=True)
